"""Deepagents module."""

from __future__ import annotations

import logging
from collections.abc import AsyncIterator, Callable
from typing import Any

from cognitia.runtime.deepagents_builtins import (
    DEEPAGENTS_NATIVE_BUILTIN_TOOLS,
    build_native_notice,
    build_portable_notice,
    filter_native_builtin_tools,
    split_native_builtin_tools,
)
from cognitia.runtime.deepagents_hitl import validate_hitl_config
from cognitia.runtime.deepagents_langchain import (
    build_langchain_messages,
    check_langchain_available,
    stream_langchain_runtime_events,
)
from cognitia.runtime.deepagents_memory import (
    build_native_invocation,
    build_native_state_notice,
    validate_native_state_config,
)
from cognitia.runtime.deepagents_models import (
    DeepAgentsModelError,
    build_deepagents_chat_model,
)
from cognitia.runtime.deepagents_native import (
    build_deepagents_graph,
    stream_deepagents_graph_events,
    validate_native_backend_config,
)
from cognitia.runtime.deepagents_tools import create_langchain_tool
from cognitia.runtime.mcp_bridge import McpBridge
from cognitia.runtime.thin.mcp_client import parse_mcp_tool_name
from cognitia.runtime.structured_output import (
    append_structured_output_instruction,
    extract_structured_output,
)
from cognitia.runtime.types import (
    Message,
    RuntimeConfig,
    RuntimeErrorData,
    RuntimeEvent,
    ToolSpec,
    TurnMetrics,
)

_log = logging.getLogger(__name__)

_DEEPAGENTS_BUILTIN_TOOLS = DEEPAGENTS_NATIVE_BUILTIN_TOOLS
__all__ = ["DeepAgentsRuntime", "_check_langchain_available", "create_langchain_tool"]


def _check_langchain_available() -> RuntimeErrorData | None:
    """Check langchain available."""
    return check_langchain_available()


class DeepAgentsRuntime:
    """Deep Agents Runtime implementation."""

    def __init__(
        self,
        config: RuntimeConfig | None = None,
        tool_executors: dict[str, Callable] | None = None,
        mcp_servers: dict[str, Any] | None = None,
    ) -> None:
        """Init."""
        self._config = config or RuntimeConfig(runtime_name="deepagents")
        self._tool_executors = tool_executors or {}
        self._mcp_bridge = McpBridge(mcp_servers) if mcp_servers else None

    async def run(
        self,
        *,
        messages: list[Message],
        system_prompt: str,
        active_tools: list[ToolSpec],
        config: RuntimeConfig | None = None,
        mode_hint: str | None = None,
    ) -> AsyncIterator[RuntimeEvent]:
        """Run."""
        dep_error = _check_langchain_available()
        if dep_error:
            yield RuntimeEvent.error(dep_error)
            return

        effective_config = config or self._config
        effective_system_prompt = append_structured_output_instruction(
            system_prompt,
            effective_config.output_format,
        )

        requested_tools = list(active_tools)
        selected_tools = self.select_active_tools(
            active_tools,
            feature_mode=effective_config.feature_mode,
            allow_native_features=effective_config.allow_native_features,
        )

        # Discover MCP tools and merge
        if self._mcp_bridge is not None:
            try:
                mcp_tools = await self._mcp_bridge.discover_all_tools()
                selected_tools = list(selected_tools) + mcp_tools
                # Register MCP executors
                for spec in mcp_tools:
                    parsed = parse_mcp_tool_name(spec.name)
                    if parsed is not None:
                        server_id, tool_name = parsed
                        self._tool_executors[spec.name] = self._mcp_bridge.create_tool_executor(
                            server_id, tool_name
                        )
            except Exception as exc:
                _log.warning("MCP discovery failed, tools unavailable: %s", exc)

        full_text = ""
        tool_calls: list[dict[str, Any]] = []
        new_messages: list[Message] = []

        try:
            stream = self._stream_langchain
            final_session_id = None
            final_native_metadata = None
            if self._should_use_native_path(effective_config):
                state_error = validate_native_state_config(effective_config.native_config)
                if state_error:
                    yield RuntimeEvent.error(state_error)
                    return

                hitl_error = validate_hitl_config(effective_config.native_config)
                if hitl_error:
                    yield RuntimeEvent.error(hitl_error)
                    return

                native_selection = split_native_builtin_tools(selected_tools)
                backend_error = validate_native_backend_config(
                    native_tool_names=native_selection.native_tool_names,
                    native_config=effective_config.native_config,
                )
                if backend_error:
                    yield RuntimeEvent.error(backend_error)
                    return

                notice = build_native_notice(
                    requested_tools,
                    feature_mode=effective_config.feature_mode,
                )
                if notice:
                    yield RuntimeEvent.status(notice)
                selected_tools = native_selection.custom_tools
                lc_messages = self._build_lc_messages(
                    messages,
                    effective_system_prompt,
                    include_system_prompt=False,
                )
                input_payload, run_config, final_native_metadata = build_native_invocation(
                    messages=lc_messages,
                    native_config=effective_config.native_config,
                )
                final_session_id = final_native_metadata.get("thread_id")
                state_notice = build_native_state_notice(final_native_metadata)
                if state_notice:
                    yield RuntimeEvent.native_notice(
                        state_notice,
                        metadata=final_native_metadata,
                    )
                stream = self._stream_native
            else:
                notice = build_portable_notice(requested_tools)
                if notice:
                    yield RuntimeEvent.status(notice)

            async for event in stream(
                messages=messages,
                system_prompt=effective_system_prompt,
                tools=selected_tools,
                model=effective_config.model,
                base_url=effective_config.base_url,
                native_config=effective_config.native_config,
                input_payload=locals().get("input_payload"),
                run_config=locals().get("run_config"),
            ):
                yield event
                if event.type == "assistant_delta":
                    full_text += str(event.data.get("text", ""))
                elif event.type == "tool_call_started":
                    tool_name = str(event.data.get("name", ""))
                    correlation_id = str(event.data.get("correlation_id", ""))
                    tool_args = event.data.get("args", {})
                    new_messages.append(
                        Message(
                            role="assistant",
                            content="",
                            tool_calls=[
                                {
                                    "id": correlation_id,
                                    "name": tool_name,
                                    "args": tool_args if isinstance(tool_args, dict) else {},
                                    "type": "tool_call",
                                }
                            ],
                            metadata={
                                "correlation_id": correlation_id,
                                "tool_name": tool_name,
                            },
                        )
                    )
                elif event.type == "tool_call_finished":
                    tool_calls.append(event.data)
                    new_messages.append(
                        Message(
                            role="tool",
                            content=str(event.data.get("result_summary", "")),
                            name=str(event.data.get("name", "")) or None,
                            metadata={
                                "correlation_id": str(event.data.get("correlation_id", "")),
                            },
                        )
                    )

        except DeepAgentsModelError as e:
            yield RuntimeEvent.error(e.error)
            return
        except Exception as e:
            yield RuntimeEvent.error(
                RuntimeErrorData(
                    kind="runtime_crash",
                    message=f"Ошибка LangChain: {e}",
                    recoverable=False,
                )
            )
            return


        new_messages.append(Message(role="assistant", content=full_text))

        yield RuntimeEvent.final(
            text=full_text,
            new_messages=new_messages,
            metrics=TurnMetrics(
                model=effective_config.model,
                tool_calls_count=len(tool_calls),
            ),
            session_id=final_session_id,
            structured_output=extract_structured_output(
                full_text,
                effective_config.output_format,
            ),
            native_metadata=final_native_metadata,
        )

    def _build_lc_messages(
        self,
        messages: list[Message],
        system_prompt: str,
        *,
        include_system_prompt: bool = True,
    ) -> list[Any]:
        """Build lc messages."""
        return build_langchain_messages(
            messages,
            system_prompt,
            include_system_prompt=include_system_prompt,
        )

    def _build_llm(self, model: str, base_url: str | None = None) -> Any:
        """Create provider-specific chat model via separate resolver."""
        return build_deepagents_chat_model(model, base_url=base_url)

    def _should_use_native_path(self, config: RuntimeConfig) -> bool:
        """Should use native path."""
        return config.is_native_mode

    async def _stream_native(
        self,
        messages: list[Message],
        system_prompt: str,
        tools: list[ToolSpec],
        model: str,
        base_url: str | None = None,
        native_config: dict[str, Any] | None = None,
        input_payload: Any = None,
        run_config: dict[str, Any] | None = None,
    ) -> AsyncIterator[RuntimeEvent]:
        """Stream native."""
        native_config = native_config or {}
        graph = build_deepagents_graph(
            model=model,
            system_prompt=system_prompt,
            tools=tools,
            tool_executors=self._tool_executors,
            base_url=base_url,
            interrupt_on=native_config.get("interrupt_on"),
            checkpointer=native_config.get("checkpointer"),
            store=native_config.get("store"),
            backend=native_config.get("backend"),
            memory=native_config.get("memory"),
            subagents=native_config.get("subagents"),
            skills=native_config.get("skills"),
            middleware=native_config.get("middleware"),
            agent_name=native_config.get("agent_name"),
        )
        async for event in stream_deepagents_graph_events(
            graph=graph,
            input_payload=input_payload or {"messages": []},
            run_config=run_config,
        ):
            yield event

    async def _stream_langchain(
        self,
        messages: list[Message],
        system_prompt: str,
        tools: list[ToolSpec],
        model: str,
        base_url: str | None = None,
        native_config: dict[str, Any] | None = None,
        input_payload: Any = None,
        run_config: dict[str, Any] | None = None,
    ) -> AsyncIterator[RuntimeEvent]:
        """Stream langchain."""
        lc_messages = self._build_lc_messages(messages, system_prompt)
        lc_tools = [
            create_langchain_tool(spec, self._tool_executors.get(spec.name)) for spec in tools
        ]
        llm = self._build_llm(model, base_url)
        runnable = llm.bind_tools(lc_tools) if lc_tools else llm

        async for event in stream_langchain_runtime_events(
            runnable=runnable,
            lc_messages=lc_messages,
        ):
            yield event

    async def cleanup(self) -> None:
        """Cleanup."""

    @staticmethod
    def filter_builtin_tools(tools: list[ToolSpec]) -> list[ToolSpec]:
        """Filter builtin tools."""
        return filter_native_builtin_tools(tools)

    @staticmethod
    def select_active_tools(
        tools: list[ToolSpec],
        *,
        feature_mode: str,
        allow_native_features: bool = False,
    ) -> list[ToolSpec]:
        """Select active tools."""
        if allow_native_features or feature_mode in {"hybrid", "native_first"}:
            return list(tools)
        return DeepAgentsRuntime.filter_builtin_tools(tools)



"""FilesystemTodoProvider - persistent todos via the filesystem.

Stores todos in {root}/{user_id}/{topic_id}/todos.json.
"""

from __future__ import annotations

import json
import os
from datetime import UTC, datetime
from pathlib import Path

from cognitia.todo.types import TodoItem


def _parse_dt(value: str | datetime) -> datetime:
    """Parse a datetime from JSON (ISO string) or pass through."""
    if isinstance(value, datetime):
        return value
    if not value:
        return datetime.now(tz=UTC)
    try:
        return datetime.fromisoformat(value)
    except (ValueError, TypeError):
        return datetime.now(tz=UTC)


class FilesystemTodoProvider:
    """TodoProvider via a JSON file."""

    def __init__(self, root_path: Path, user_id: str, topic_id: str, max_todos: int = 100) -> None:
        self._max_todos = max_todos
        self._file = Path(root_path) / user_id / topic_id / "todos.json"

    async def read_todos(self) -> list[TodoItem]:
        """Read todos from the file."""
        if not self._file.exists():
            return []
        raw = json.loads(self._file.read_text(encoding="utf-8"))
        return [
            TodoItem(
                id=t["id"],
                content=t["content"],
                status=t["status"],
                created_at=_parse_dt(t.get("created_at", "")),
                updated_at=_parse_dt(t.get("updated_at", "")),
            )
            for t in raw
        ]

    async def write_todos(self, todos: list[TodoItem]) -> None:
        """Write todos to the file (bulk replace)."""
        if len(todos) > self._max_todos:
            msg = f"Превышен лимит max_todos ({self._max_todos})"
            raise ValueError(msg)
        self._file.parent.mkdir(parents=True, exist_ok=True)
        data = [t.to_dict() for t in todos]
        tmp = self._file.with_suffix(".json.tmp")
        try:
            tmp.write_text(json.dumps(data, ensure_ascii=False, indent=2), encoding="utf-8")
            os.replace(str(tmp), str(self._file))
        except Exception:
            tmp.unlink(missing_ok=True)
            raise



"""InMemoryTodoProvider - session-scoped todo storage.

Data lives in process memory. Multi-tenant: each instance
is created for a specific user_id + topic_id.
"""

from __future__ import annotations

from cognitia.todo.types import TodoItem


class InMemoryTodoProvider:
    """In-memory todo storage (session-scoped).

    Each instance is a separate namespace (user_id + topic_id).
    """

    def __init__(
        self,
        user_id: str,
        topic_id: str,
        max_todos: int = 100,
    ) -> None:
        self._user_id = user_id
        self._topic_id = topic_id
        self._max_todos = max_todos
        self._todos: list[TodoItem] = []

    async def read_todos(self) -> list[TodoItem]:
        """Read all todos."""
        return list(self._todos)

    async def write_todos(self, todos: list[TodoItem]) -> None:
        """Write todos (bulk replace)."""
        if len(todos) > self._max_todos:
            msg = f"Превышен лимит max_todos ({self._max_todos}), передано {len(todos)}"
            raise ValueError(msg)
        self._todos = list(todos)



"""RAG / Retriever Protocol — document retrieval and context injection.

Provides:
- Document: frozen dataclass for retrieved documents
- Retriever: runtime-checkable protocol for retrieval backends
- SimpleRetriever: builtin in-memory retriever using word overlap scoring
- RagInputFilter: InputFilter that injects retrieved context into system_prompt
"""

from __future__ import annotations

from collections import Counter
from dataclasses import dataclass, field
from typing import Any, Protocol, runtime_checkable

from cognitia.runtime.types import Message


@dataclass(frozen=True)
class Document:
    """A retrieved document with content, metadata, and optional relevance score."""

    content: str
    metadata: dict[str, Any] = field(default_factory=dict)
    score: float | None = None


@runtime_checkable
class Retriever(Protocol):
    """Protocol for document retrieval backends.

    Implementations must provide an async retrieve method that returns
    documents ranked by relevance to the query.
    """

    async def retrieve(self, query: str, top_k: int = 5) -> list[Document]:
        """Retrieve top_k documents relevant to query."""
        ...


class SimpleRetriever:
    """In-memory retriever using word overlap scoring.

    Intended for development and testing. Uses simple word overlap
    (intersection of word counters) as a proxy for TF-IDF relevance.

    Args:
        documents: List of documents to search over.
    """

    def __init__(self, documents: list[Document]) -> None:
        self._documents = documents

    async def retrieve(self, query: str, top_k: int = 5) -> list[Document]:
        """Retrieve documents ranked by word overlap with query.

        Returns empty list if query is empty or no documents match.
        """
        if not query.strip():
            return []

        query_words = Counter(query.lower().split())

        scored: list[Document] = []
        for doc in self._documents:
            doc_words = Counter(doc.content.lower().split())
            # Sum of minimum counts for overlapping words
            overlap = sum(
                min(query_words[w], doc_words[w])
                for w in query_words
                if w in doc_words
            )
            if overlap > 0:
                scored.append(
                    Document(
                        content=doc.content,
                        metadata=doc.metadata,
                        score=float(overlap),
                    )
                )

        scored.sort(key=lambda d: d.score or 0.0, reverse=True)
        return scored[:top_k]


class RagInputFilter:
    """InputFilter that injects retrieved documents into system_prompt.

    Extracts the query from the last user message, retrieves relevant
    documents via the configured Retriever, and prepends them to the
    system prompt wrapped in <context>...</context> tags.

    Args:
        retriever: A Retriever implementation.
        top_k: Maximum number of documents to retrieve per query.
    """

    def __init__(self, retriever: Retriever, top_k: int = 5) -> None:
        self._retriever = retriever
        self._top_k = top_k

    async def filter(
        self, messages: list[Message], system_prompt: str
    ) -> tuple[list[Message], str]:
        """Retrieve documents and inject them into system_prompt.

        Returns (messages, enriched_system_prompt). Messages are never modified.
        """
        query = self._extract_last_user_text(messages)
        if not query:
            return messages, system_prompt

        docs = await self._retriever.retrieve(query, self._top_k)
        if not docs:
            return messages, system_prompt

        context_block = self._format_context(docs)
        enriched = f"{context_block}\n{system_prompt}" if system_prompt else context_block
        return messages, enriched

    @staticmethod
    def _extract_last_user_text(messages: list[Message]) -> str:
        """Extract text from the last user message."""
        for msg in reversed(messages):
            if msg.role == "user" and msg.content:
                return msg.content
        return ""

    @staticmethod
    def _format_context(docs: list[Document]) -> str:
        """Format documents as a <context> block."""
        lines = [doc.content for doc in docs]
        inner = "\n".join(lines)
        return f"<context>\n{inner}\n</context>"



"""Types for team orchestration - multi-agent teams.

TeamConfig, TeamStatus, TeamMessage, InternalTeamState, compose_worker_task.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, Literal

from cognitia.orchestration.subagent_types import SubagentSpec, SubagentStatus

if TYPE_CHECKING:
    from cognitia.orchestration.message_bus import MessageBus

TeamState = Literal["idle", "running", "completed", "failed"]


@dataclass(frozen=True)
class TeamConfig:
    """Configuration teams agents."""

    lead_prompt: str
    worker_specs: list[SubagentSpec]
    max_workers: int = 4
    communication: Literal["message_passing", "shared_state"] = "message_passing"


@dataclass(frozen=True)
class TeamMessage:
    """Message between agentamand and toomande."""

    from_agent: str
    to_agent: str
    content: str
    timestamp: datetime


@dataclass(frozen=True)
class TeamStatus:
    """Team Status status object."""

    team_id: str
    state: TeamState = "idle"
    workers: dict[str, SubagentStatus] = field(default_factory=dict)
    messages_exchanged: int = 0


class InternalTeamState:
    """Internal Team State implementation."""

    def __init__(
        self,
        *,
        config: TeamConfig,
        worker_ids: dict[str, str],
        started_at: datetime,
        bus: MessageBus,
        task: str,
    ) -> None:
        self.config = config
        self.worker_ids = worker_ids
        self.started_at = started_at
        self.bus = bus
        self.task = task
        self.paused_workers: set[str] = set()


def compose_worker_task(*, config: TeamConfig, worker_name: str, task: str) -> str:
    """Compose worker task."""
    return (
        f"{config.lead_prompt}\n\n"
        f"Ты worker '{worker_name}'.\n"
        f"Выполни подзадачу в контексте общей цели:\n{task}"
    )



"""UI Event Projection — transforms RuntimeEvent streams into UI state.

Provides a structured way to project a stream of RuntimeEvent into
a renderable UIState suitable for chat UIs, dashboards, or logs.

Architecture:
- UIBlock union: TextBlock | ToolCallBlock | ToolResultBlock | ErrorBlock
- UIMessage: role + ordered list of UIBlocks + optional timestamp
- UIState: full conversation state (messages + status + metadata)
- EventProjection protocol: apply(event) -> UIState
- ChatProjection: built-in implementation for chat-style UIs
- project_stream: async helper that yields UIState per event
"""

from __future__ import annotations

import time
from collections.abc import AsyncIterator
from dataclasses import dataclass, field
from typing import Any, Protocol, Union, runtime_checkable

from cognitia.runtime.types import RuntimeEvent


# ---------------------------------------------------------------------------
# UIBlock — union of block types
# ---------------------------------------------------------------------------


@dataclass(frozen=True)
class TextBlock:
    """Accumulated text content from assistant."""

    text: str


@dataclass(frozen=True)
class ToolCallBlock:
    """A tool call initiation."""

    name: str
    args: dict[str, Any]
    correlation_id: str


@dataclass(frozen=True)
class ToolResultBlock:
    """A tool call result."""

    name: str
    ok: bool
    summary: str
    correlation_id: str


@dataclass(frozen=True)
class ErrorBlock:
    """An error that occurred during processing."""

    kind: str
    message: str


UIBlock = Union[TextBlock, ToolCallBlock, ToolResultBlock, ErrorBlock]


# ---------------------------------------------------------------------------
# UIMessage & UIState
# ---------------------------------------------------------------------------

_BLOCK_TYPE_MAP = {
    "text": TextBlock,
    "tool_call": ToolCallBlock,
    "tool_result": ToolResultBlock,
    "error": ErrorBlock,
}


def _block_to_dict(block: UIBlock) -> dict[str, Any]:
    """Serialize a UIBlock to a dict with a 'type' discriminator."""
    if isinstance(block, TextBlock):
        return {"type": "text", "text": block.text}
    if isinstance(block, ToolCallBlock):
        return {
            "type": "tool_call",
            "name": block.name,
            "args": block.args,
            "correlation_id": block.correlation_id,
        }
    if isinstance(block, ToolResultBlock):
        return {
            "type": "tool_result",
            "name": block.name,
            "ok": block.ok,
            "summary": block.summary,
            "correlation_id": block.correlation_id,
        }
    if isinstance(block, ErrorBlock):
        return {"type": "error", "kind": block.kind, "message": block.message}
    raise ValueError(f"Unknown block type: {type(block)}")  # pragma: no cover


def _block_from_dict(d: dict[str, Any]) -> UIBlock:
    """Deserialize a UIBlock from a dict with 'type' discriminator."""
    block_type = d["type"]
    if block_type == "text":
        return TextBlock(text=d["text"])
    if block_type == "tool_call":
        return ToolCallBlock(
            name=d["name"], args=d["args"], correlation_id=d["correlation_id"]
        )
    if block_type == "tool_result":
        return ToolResultBlock(
            name=d["name"],
            ok=d["ok"],
            summary=d["summary"],
            correlation_id=d["correlation_id"],
        )
    if block_type == "error":
        return ErrorBlock(kind=d["kind"], message=d["message"])
    raise ValueError(f"Unknown block type: {block_type}")  # pragma: no cover


@dataclass
class UIMessage:
    """A single message in the UI conversation."""

    role: str
    blocks: list[UIBlock]
    timestamp: float | None = None


@dataclass
class UIState:
    """Full UI conversation state — serializable snapshot."""

    messages: list[UIMessage]
    status: str = "idle"
    metadata: dict[str, Any] = field(default_factory=dict)

    def to_dict(self) -> dict[str, Any]:
        """Serialize to a JSON-compatible dict."""
        return {
            "messages": [
                {
                    "role": msg.role,
                    "blocks": [_block_to_dict(b) for b in msg.blocks],
                    "timestamp": msg.timestamp,
                }
                for msg in self.messages
            ],
            "status": self.status,
            "metadata": self.metadata,
        }

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> UIState:
        """Deserialize from a dict produced by to_dict()."""
        messages = [
            UIMessage(
                role=m["role"],
                blocks=[_block_from_dict(b) for b in m["blocks"]],
                timestamp=m.get("timestamp"),
            )
            for m in d["messages"]
        ]
        return cls(
            messages=messages,
            status=d.get("status", "idle"),
            metadata=d.get("metadata", {}),
        )


# ---------------------------------------------------------------------------
# EventProjection Protocol
# ---------------------------------------------------------------------------


@runtime_checkable
class EventProjection(Protocol):
    """Protocol for projecting RuntimeEvent streams into UIState."""

    def apply(self, event: RuntimeEvent) -> UIState:
        """Apply a single event and return the updated UIState."""
        ...  # pragma: no cover


# ---------------------------------------------------------------------------
# ChatProjection — built-in implementation
# ---------------------------------------------------------------------------


class ChatProjection:
    """Built-in EventProjection for chat-style UIs.

    Maintains internal mutable UIState and updates it on each apply() call.
    All assistant content (text deltas, tool calls/results, errors) is
    accumulated into assistant-role UIMessages.
    """

    def __init__(self) -> None:
        self._state = UIState(messages=[])

    @property
    def state(self) -> UIState:
        """Current UI state snapshot."""
        return self._state

    def apply(self, event: RuntimeEvent) -> UIState:
        """Apply a RuntimeEvent and return updated UIState."""
        handler = _EVENT_HANDLERS.get(event.type)
        if handler is not None:
            handler(self, event)
        return self._state

    def _ensure_assistant_message(self) -> UIMessage:
        """Get or create the current assistant message."""
        if not self._state.messages or self._state.messages[-1].role != "assistant":
            msg = UIMessage(role="assistant", blocks=[], timestamp=time.time())
            self._state.messages.append(msg)
        return self._state.messages[-1]

    def _handle_assistant_delta(self, event: RuntimeEvent) -> None:
        msg = self._ensure_assistant_message()
        text = event.data.get("text", "")
        if msg.blocks and isinstance(msg.blocks[-1], TextBlock):
            # Accumulate into existing TextBlock (frozen, so replace)
            old = msg.blocks[-1]
            msg.blocks[-1] = TextBlock(text=old.text + text)
        else:
            msg.blocks.append(TextBlock(text=text))

    def _handle_tool_call_started(self, event: RuntimeEvent) -> None:
        msg = self._ensure_assistant_message()
        msg.blocks.append(
            ToolCallBlock(
                name=event.data.get("name", ""),
                args=event.data.get("args", {}),
                correlation_id=event.data.get("correlation_id", ""),
            )
        )

    def _handle_tool_call_finished(self, event: RuntimeEvent) -> None:
        msg = self._ensure_assistant_message()
        msg.blocks.append(
            ToolResultBlock(
                name=event.data.get("name", ""),
                ok=event.data.get("ok", False),
                summary=event.data.get("result_summary", ""),
                correlation_id=event.data.get("correlation_id", ""),
            )
        )

    def _handle_error(self, event: RuntimeEvent) -> None:
        msg = self._ensure_assistant_message()
        msg.blocks.append(
            ErrorBlock(
                kind=event.data.get("kind", "runtime_crash"),
                message=event.data.get("message", "Unknown error"),
            )
        )

    def _handle_status(self, event: RuntimeEvent) -> None:
        self._state.status = event.data.get("text", "")

    def _handle_final(self, event: RuntimeEvent) -> None:
        self._state.status = "done"
        # Propagate useful metadata from final event
        for key in ("session_id", "total_cost_usd", "metrics", "usage"):
            if key in event.data:
                self._state.metadata[key] = event.data[key]


# Handler dispatch table (avoids if/elif chain, OCP-friendly)
_EVENT_HANDLERS: dict[str, Any] = {
    "assistant_delta": ChatProjection._handle_assistant_delta,
    "tool_call_started": ChatProjection._handle_tool_call_started,
    "tool_call_finished": ChatProjection._handle_tool_call_finished,
    "error": ChatProjection._handle_error,
    "status": ChatProjection._handle_status,
    "final": ChatProjection._handle_final,
}


# ---------------------------------------------------------------------------
# project_stream — async helper
# ---------------------------------------------------------------------------


async def project_stream(
    events: AsyncIterator[RuntimeEvent],
    projection: EventProjection,
) -> AsyncIterator[UIState]:
    """Project an async stream of RuntimeEvent into UIState snapshots.

    Yields a UIState after each event is applied.
    """
    async for event in events:
        yield projection.apply(event)



"""CognitiaStack - facade for quick initialization of library components.

Single assembly point: the application passes config/prompts/skills paths
and capability configs, then receives ready-to-wire components.
"""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, field
from importlib.resources import files
from pathlib import Path
from typing import Any

from cognitia.config.role_router import RoleRouterConfig, load_role_router_config
from cognitia.config.role_skills import YamlRoleSkillsLoader
from cognitia.context import DefaultContextBuilder
from cognitia.policy import DefaultToolIdCodec, DefaultToolPolicy
from cognitia.policy.tool_selector import ToolBudgetConfig
from cognitia.protocols import LocalToolResolver
from cognitia.routing import KeywordRoleRouter
from cognitia.runtime.factory import RuntimeFactory
from cognitia.runtime.model_policy import ModelPolicy
from cognitia.runtime.types import RuntimeConfig, ToolSpec
from cognitia.skills import SkillRegistry
from cognitia.skills.loader import YamlSkillLoader


@dataclass
class CognitiaStack:
    """A ready-made set of library components for the application.

    Created via CognitiaStack.create() - the single factory.
    Capability tools are assembled automatically and exposed through capability_specs/executors.
    """

    skill_registry: SkillRegistry
    context_builder: DefaultContextBuilder
    role_skills_loader: YamlRoleSkillsLoader
    role_router: KeywordRoleRouter
    role_router_config: RoleRouterConfig
    tool_policy: DefaultToolPolicy
    tool_id_codec: DefaultToolIdCodec
    model_policy: ModelPolicy
    runtime_factory: RuntimeFactory
    runtime_config: RuntimeConfig
    local_tool_resolver: LocalToolResolver | None
    # Capability tools - merged from all enabled capabilities
    capability_specs: dict[str, ToolSpec] = field(default_factory=dict)
    capability_executors: dict[str, Callable] = field(default_factory=dict)
    memory_bank_provider: Any | None = None
    memory_bank_prompt: str | None = None
    tool_budget_config: ToolBudgetConfig | None = None

    @classmethod
    def create(
        cls,
        *,
        prompts_dir: Path,
        skills_dir: Path,
        project_root: Path,
        escalate_roles: set[str] | None = None,
        runtime_config: RuntimeConfig | None = None,
        local_tool_resolver: LocalToolResolver | None = None,
        # --- Capability configs ---
        sandbox_provider: Any | None = None,
        web_provider: Any | None = None,
        todo_provider: Any | None = None,
        memory_bank_provider: Any | None = None,
        memory_bank_prompt: str | None = None,
        plan_manager: Any | None = None,
        plan_user_id: str = "",
        plan_topic_id: str = "",
        thinking_enabled: bool = True,
        allowed_system_tools: set[str] | None = None,
        tool_budget_config: ToolBudgetConfig | None = None,
    ) -> CognitiaStack:
        """Create all library components from config paths.

        Args:
            prompts_dir: Directory with prompts.
            skills_dir: Directory with skills.
            project_root: Project root.
            escalate_roles: Roles for model escalation.
            runtime_config: Runtime config.
            local_tool_resolver: App-level resolver for local tools.
            sandbox_provider: SandboxProvider for builtin tools.
            web_provider: WebProvider for web tools.
            todo_provider: TodoProvider for todo tools.
            memory_bank_provider: MemoryBankProvider for memory tools.
            memory_bank_prompt: Prompt instruction for Memory Bank (P_MEMORY layer).
            plan_manager: PlanManager for planning tools.
            plan_user_id: user_id for the plan namespace.
            plan_topic_id: topic_id for the plan namespace.
            thinking_enabled: Enable the thinking tool.
            allowed_system_tools: Whitelist for system tools in ToolPolicy.
            tool_budget_config: Tool budget configuration.

        Returns:
            CognitiaStack with ready-to-use components.
        """
        # Skills
        skill_loader = YamlSkillLoader(skills_dir, project_root=project_root)
        loaded_skills = skill_loader.load_all()
        skill_registry = SkillRegistry(
            loaded_skills,
            settings_mcp=skill_loader.settings_mcp_servers,
        )

        # Context
        context_builder = DefaultContextBuilder(prompts_dir)

        # Role config
        role_skills_loader = YamlRoleSkillsLoader(
            prompts_dir / "role_skills.yaml",
        )
        router_config = load_role_router_config(
            prompts_dir / "role_router.yaml",
        )
        role_router = KeywordRoleRouter(
            default_role=router_config.default_role,
            keyword_map=router_config.keywords,
        )

        # Policy - with system tool whitelist
        tool_id_codec = DefaultToolIdCodec()
        tool_policy = DefaultToolPolicy(
            codec=tool_id_codec,
            allowed_system_tools=allowed_system_tools,
        )

        # Model policy
        model_policy = ModelPolicy(
            escalate_roles=escalate_roles or set(),
        )

        # Runtime factory
        runtime_factory = RuntimeFactory()
        runtime_cfg = runtime_config or RuntimeConfig()

        # Capability tools - collect from enabled capabilities
        from cognitia.bootstrap.capabilities import collect_capability_tools

        cap_specs, cap_executors = collect_capability_tools(
            sandbox_provider=sandbox_provider,
            web_provider=web_provider,
            todo_provider=todo_provider,
            memory_bank_provider=memory_bank_provider,
            plan_manager=plan_manager,
            plan_user_id=plan_user_id,
            plan_topic_id=plan_topic_id,
            thinking_enabled=thinking_enabled,
            tool_budget_config=tool_budget_config,
        )

        if memory_bank_prompt is None and memory_bank_provider is not None:
            default_prompt_path = files("cognitia.memory_bank").joinpath("default_prompt.md")
            try:
                memory_bank_prompt = default_prompt_path.read_text(encoding="utf-8")
            except OSError:
                memory_bank_prompt = None

        return cls(
            skill_registry=skill_registry,
            context_builder=context_builder,
            role_skills_loader=role_skills_loader,
            role_router=role_router,
            role_router_config=router_config,
            tool_policy=tool_policy,
            tool_id_codec=tool_id_codec,
            model_policy=model_policy,
            runtime_factory=runtime_factory,
            runtime_config=runtime_cfg,
            local_tool_resolver=local_tool_resolver,
            capability_specs=cap_specs,
            capability_executors=cap_executors,
            memory_bank_provider=memory_bank_provider,
            memory_bank_prompt=memory_bank_prompt,
            tool_budget_config=tool_budget_config,
        )



"""Multi-agent protocols -- agent-as-tool, task queue, and agent registry contracts.

ISP-compliant: AgentTool has 1 method, TaskQueue has 5 methods, AgentRegistry has 5 methods.
Dependencies: cognitia domain types and stdlib only.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Protocol, runtime_checkable

from cognitia.multi_agent.task_types import TaskFilter, TaskItem

if TYPE_CHECKING:
    from cognitia.multi_agent.registry_types import AgentFilter, AgentRecord, AgentStatus
    from cognitia.runtime.types import ToolSpec


@runtime_checkable
class AgentTool(Protocol):
    """Port: agent exposable as a tool for other agents.

    Any runtime or agent wrapper implementing this Protocol
    can be used as a sub-agent tool in another agent's tool list.
    """

    def as_tool(self, name: str, description: str) -> ToolSpec: ...


@runtime_checkable
class TaskQueue(Protocol):
    """Async task queue for multi-agent coordination.

    Exactly 5 methods (ISP limit). All methods are async.
    """

    async def put(self, item: TaskItem) -> None:
        """Add a task to the queue."""
        ...

    async def get(self, filters: TaskFilter | None = None) -> TaskItem | None:
        """Claim the highest-priority TODO task matching filters.

        Implementations must return the claimed task in IN_PROGRESS status.
        Returns None if no matching task is available.
        """
        ...

    async def complete(self, task_id: str) -> bool:
        """Mark a task as done. Returns True if found and completed."""
        ...

    async def cancel(self, task_id: str) -> bool:
        """Mark a task as cancelled. Returns True if found and cancelled."""
        ...

    async def list_tasks(
        self, filters: TaskFilter | None = None
    ) -> list[TaskItem]:
        """List all tasks matching filters. Returns all if filters is None."""
        ...


@runtime_checkable
class AgentRegistry(Protocol):
    """Port: registry for managing agent lifecycle records.

    Provides CRUD operations for AgentRecord instances with
    filtering support. Exactly 5 methods (ISP limit).
    """

    async def register(self, record: AgentRecord) -> None: ...

    async def get(self, agent_id: str) -> AgentRecord | None: ...

    async def list_agents(
        self, filters: AgentFilter | None = None,
    ) -> list[AgentRecord]: ...

    async def update_status(self, agent_id: str, status: AgentStatus) -> bool: ...

    async def remove(self, agent_id: str) -> bool: ...



"""EventBus — publish-subscribe event bus for runtime events.

Provides:
- EventBus: Protocol defining the pub-sub contract.
- InMemoryEventBus: Default fire-and-forget implementation.
"""

from __future__ import annotations

import asyncio
from typing import Any, Callable, Protocol, runtime_checkable


@runtime_checkable
class EventBus(Protocol):
    """Publish-subscribe event bus for runtime events."""

    def subscribe(self, event_type: str, callback: Callable[..., Any]) -> str:
        """Register a callback for event_type. Returns subscription ID."""
        ...

    def unsubscribe(self, subscription_id: str) -> None:
        """Remove a subscription by ID."""
        ...

    async def emit(self, event_type: str, data: dict[str, Any]) -> None:
        """Emit an event to all subscribers of event_type."""
        ...


class InMemoryEventBus:
    """Default event bus -- fire-and-forget async callbacks.

    Callbacks that raise exceptions are silently ignored (fire-and-forget).
    Supports both sync and async callbacks.
    """

    def __init__(self) -> None:
        self._subscribers: dict[str, dict[str, Callable[..., Any]]] = {}
        self._counter = 0

    def subscribe(self, event_type: str, callback: Callable[..., Any]) -> str:
        """Register a callback for event_type. Returns unique subscription ID."""
        sub_id = f"sub_{self._counter}"
        self._counter += 1
        self._subscribers.setdefault(event_type, {})[sub_id] = callback
        return sub_id

    def unsubscribe(self, subscription_id: str) -> None:
        """Remove a subscription by ID. No-op if ID not found."""
        for subs in self._subscribers.values():
            subs.pop(subscription_id, None)

    async def emit(self, event_type: str, data: dict[str, Any]) -> None:
        """Emit an event to all subscribers. Errors in callbacks are swallowed."""
        for cb in list(self._subscribers.get(event_type, {}).values()):
            try:
                result = cb(data)
                if asyncio.iscoroutine(result):
                    await result
            except Exception:  # noqa: BLE001
                pass  # fire-and-forget



"""HITL/interrupt helpers for DeepAgents native path."""

from __future__ import annotations

from typing import Any

from cognitia.runtime.types import RuntimeErrorData, RuntimeEvent


def validate_hitl_config(
    native_config: dict[str, Any],
) -> RuntimeErrorData | None:
    """Validate hitl config."""
    if native_config.get("interrupt_on") and native_config.get("checkpointer") is None:
        return RuntimeErrorData(
            kind="capability_unsupported",
            message=(
                "DeepAgents interrupt_on требует checkpointer в native_config, "
                "иначе resume/HITL невозможны."
            ),
            recoverable=False,
        )
    return None


def build_interrupt_events(chunk: dict[str, Any]) -> list[RuntimeEvent]:
    """Build interrupt events."""
    interrupts = chunk.get("__interrupt__") or ()
    events: list[RuntimeEvent] = []

    for raw_interrupt in interrupts:
        value = _interrupt_value(raw_interrupt)
        interrupt_id = _interrupt_id(raw_interrupt)

        if isinstance(value, dict) and {
            "action_requests",
            "review_configs",
        }.issubset(value):
            events.extend(
                _approval_events_from_hitl_request(
                    value,
                    interrupt_id=interrupt_id,
                )
            )
        elif isinstance(value, str):
            events.append(
                RuntimeEvent.user_input_requested(
                    prompt=value,
                    interrupt_id=interrupt_id,
                )
            )
        else:
            events.append(
                RuntimeEvent.native_notice(
                    "DeepAgents emitted unsupported interrupt payload",
                    metadata={
                        "interrupt_id": interrupt_id,
                        "value": repr(value),
                    },
                )
            )

    return events


def _approval_events_from_hitl_request(
    request: dict[str, Any],
    *,
    interrupt_id: str | None,
) -> list[RuntimeEvent]:
    action_requests = [
        item for item in (request.get("action_requests") or []) if isinstance(item, dict)
    ]
    review_configs = [
        item for item in (request.get("review_configs") or []) if isinstance(item, dict)
    ]
    events: list[RuntimeEvent] = []

    total = max(len(action_requests), len(review_configs))
    for idx in range(total):
        action: dict[str, Any] = action_requests[idx] if idx < len(action_requests) else {}
        review: dict[str, Any] = review_configs[idx] if idx < len(review_configs) else {}
        action_name = action.get("name") or review.get("action_name") or ""
        if not action_name:
            events.append(
                RuntimeEvent.native_notice(
                    "DeepAgents interrupt payload missing action name",
                    metadata={"interrupt_id": interrupt_id},
                )
            )
            continue

        events.append(
            RuntimeEvent.approval_required(
                action_name=action_name,
                args=dict(action.get("args") or {}),
                allowed_decisions=list(review.get("allowed_decisions") or []),
                interrupt_id=interrupt_id,
                description=str(action.get("description") or ""),
            )
        )

    return events


def _interrupt_value(raw_interrupt: Any) -> Any:
    if isinstance(raw_interrupt, dict):
        return raw_interrupt.get("value")
    return getattr(raw_interrupt, "value", None)


def _interrupt_id(raw_interrupt: Any) -> str | None:
    if isinstance(raw_interrupt, dict):
        return raw_interrupt.get("id")
    return getattr(raw_interrupt, "id", None)



"""Tool protocols -- tool ID encoding and local tool resolution."""

from __future__ import annotations

from typing import Any, Protocol


class ToolIdCodec(Protocol):
    """Port: tool name normalization."""

    def matches(self, tool_name: str, server_id: str) -> bool: ...

    def encode(self, server_id: str, tool_name: str) -> str: ...

    def extract_server(self, tool_name: str) -> str | None: ...


class LocalToolResolver(Protocol):
    """Port: local tool resolver (ISP, 2 methods).

    Application implements this Protocol so the library can obtain
    a callable by tool_name. Library has no knowledge of concrete tools.
    """

    def resolve(self, tool_name: str) -> Any | None:
        """Get callable for a local tool by name.

        Returns:
            Callable or None if tool not found.
        """
        ...

    def list_tools(self) -> list[str]:
        """List available local tools."""
        ...



"""Tavily search provider - AI-optimized search.

Requires an API key (TAVILY_API_KEY). Free tier: 1000 req/month.
Optional dependency: tavily-python (pip install cognitia[web-tavily]).
"""

from __future__ import annotations

import asyncio

import structlog

from cognitia.tools.web_protocols import SearchResult

try:
    from tavily import TavilyClient  # type: ignore[import-untyped]
except ImportError:
    TavilyClient = None  # type: ignore[assignment,misc]

_log = structlog.get_logger(component="web_search.tavily")


class TavilySearchProvider:
    """Search via the Tavily API (optimized for AI agents).

    Requires an API key. Falls back gracefully if tavily-python is missing.
    """

    def __init__(self, api_key: str) -> None:
        if not api_key:
            raise ValueError("TAVILY_API_KEY обязателен для TavilySearchProvider")
        self._api_key = api_key

    async def search(self, query: str, max_results: int = 5) -> list[SearchResult]:
        """Search via the Tavily API.

        Args:
            query: Search query. Empty/whitespace -> empty list.
            max_results: Maximum number of results.

        Returns:
            List of SearchResult. Empty if tavily-python is unavailable or an error occurs.
        """
        if TavilyClient is None:
            return []
        if not query or not query.strip():
            return []

        def _sync_search() -> dict:
            try:
                client = TavilyClient(api_key=self._api_key)
                return client.search(query=query, max_results=max_results)
            except Exception as exc:
                _log.warning("tavily_search_failed", query=query[:100], error=str(exc))
                return {}

        loop = asyncio.get_running_loop()
        response = await loop.run_in_executor(None, _sync_search)

        results = response.get("results", [])
        return [
            SearchResult(
                title=r.get("title", ""),
                url=r.get("url", ""),
                snippet=r.get("content", ""),
            )
            for r in results
        ]



"""Protocols for orchestration: PlanStore, PlannerMode.

ISP-joint interfaces for persistence and planoin and managed mode.
"""

from __future__ import annotations

from collections.abc import AsyncIterator
from typing import Protocol, runtime_checkable

from cognitia.orchestration.types import ApprovalSource, Plan, PlanStep


@runtime_checkable
class PlanStore(Protocol):
    """Store planoin - ISP: <=4 methods.

  Multi-tenant: user_id + topic_id in each other.
  """

    async def save(self, plan: Plan) -> None:
        """Save or update plan."""
        ...

    async def load(self, plan_id: str) -> Plan | None:
        """Load."""
        ...

    async def list_plans(self, user_id: str, topic_id: str) -> list[Plan]:
        """List plans."""
        ...

    async def update_step(self, plan_id: str, step: PlanStep) -> None:
        """Update step."""
        ...


class PlannerMode(Protocol):
    """Planner Mode protocol."""

    async def generate_plan(self, goal: str, context: str) -> Plan:
        """Generate plan via LLM."""
        ...

    async def approve(self, plan: Plan, by: ApprovalSource) -> Plan:
        """Approve plan (programmatically or via user)."""
        ...

    async def execute_step(self, plan: Plan, step_id: str) -> PlanStep:
        """Execute step."""
        ...

    def execute_all(self, plan: Plan) -> AsyncIterator[PlanStep]:
        """Execute everything step and by, therefore, the results."""
        ...

    async def replan(self, plan: Plan, feedback: str) -> Plan:
        """Regenerate plan taking into account feedback."""
        ...



"""WorkflowPipeline - generic structured pipeline protocol."""

from __future__ import annotations

from typing import Protocol

from cognitia.orchestration.types import Plan
from cognitia.orchestration.verification_types import VerificationResult


class WorkflowPipeline(Protocol):
    """Generic structured pipeline: research -> plan -> execute -> review -> verify."""

    async def research(self, goal: str) -> str: ...

    async def plan(self, research: str) -> Plan: ...

    async def execute(self, plan: Plan) -> str: ...

    async def review(self, result: str) -> str: ...

    async def verify(self, result: str) -> VerificationResult: ...



"""SDK Bridge - convert cognitia HookRegistry into SDK HookMatcher format.

Allows using cognitia HookRegistry to register hooks,
then converting them into a format understood by Claude Agent SDK.
"""

from __future__ import annotations

from typing import Any

from claude_agent_sdk import HookMatcher

from cognitia.hooks.registry import HookEntry, HookRegistry


def registry_to_sdk_hooks(
    registry: HookRegistry,
) -> dict[str, list[HookMatcher]] | None:
    """Convert cognitia HookRegistry into an SDK hooks dict.

    Returns:
        dict[HookEvent, list[HookMatcher]] for passing to ClaudeAgentOptions.hooks,
        or None if the registry is empty.
    """
    events = registry.list_events()
    if not events:
        return None

    sdk_hooks: dict[str, list[HookMatcher]] = {}

    for event_name in events:
        entries = registry.get_hooks(event_name)
        matchers = [_entry_to_matcher(entry) for entry in entries]
        sdk_hooks[event_name] = matchers

    return sdk_hooks


def _entry_to_matcher(entry: HookEntry) -> HookMatcher:
    """Convert HookEntry into an SDK HookMatcher."""
    sdk_callback = _wrap_callback(entry.callback)
    return HookMatcher(
        matcher=entry.matcher or None,
        hooks=[sdk_callback],
    )


def _wrap_callback(cognitia_callback: Any) -> Any:
    """Wrap a cognitia callback into an SDK-compatible HookCallback.

    SDK HookCallback signature: (input: HookInput, tool_use_id: str | None, context: HookContext) -> HookJSONOutput
    Cognitia callback signature: (**kwargs) -> dict | None
    """

    async def sdk_callback(
        hook_input: dict[str, Any],
        tool_use_id: str | None,
        context: dict[str, Any],
    ) -> dict[str, Any]:
        # Pass all hook_input fields as kwargs to the cognitia callback
        result = await cognitia_callback(**hook_input)
        if result is None:
            return {"continue_": True}
        return result

    return sdk_callback



"""Subagent Protocol module."""

from __future__ import annotations

from typing import Protocol, runtime_checkable

from cognitia.orchestration.subagent_types import SubagentResult, SubagentSpec, SubagentStatus


@runtime_checkable
class SubagentOrchestrator(Protocol):
    """Subagent Orchestrator protocol."""

    async def spawn(self, spec: SubagentSpec, task: str) -> str:
        """Run subagent. Returns agent_id."""
        ...

    async def get_status(self, agent_id: str) -> SubagentStatus:
        """Get status."""
        ...

    async def cancel(self, agent_id: str) -> None:
        """Cancel subagent."""
        ...

    async def wait(self, agent_id: str) -> SubagentResult:
        """Wait for the subagent to complete."""
        ...

    async def list_active(self) -> list[str]:
        """List active."""
        ...



"""DDL export for Todos - app-level migrations."""

from __future__ import annotations


def get_todo_ddl() -> list[str]:
    """Dialect-agnostic DDL for the todos table."""
    return [
        """
        CREATE TABLE IF NOT EXISTS todos (
            id TEXT PRIMARY KEY,
            user_id TEXT NOT NULL,
            topic_id TEXT NOT NULL,
            content TEXT NOT NULL,
            status TEXT NOT NULL DEFAULT 'pending',
            created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
            updated_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
        )
        """,
        "CREATE INDEX IF NOT EXISTS idx_todos_user_topic ON todos(user_id, topic_id)",
    ]



"""Prompts module."""

from __future__ import annotations

from cognitia.runtime.types import ToolSpec


def build_react_prompt(
    system_prompt: str,
    tools: list[ToolSpec],
) -> str:
    """Build react prompt."""
    tool_descs = _format_tools(tools)

    return f"""{system_prompt}

## Инструкции по формату ответа

Ты ОБЯЗАН отвечать ТОЛЬКО валидным JSON без markdown-обёртки.
Возможны два варианта:

### Вариант A — вызов инструмента:
{{"type": "tool_call", "tool": {{"name": "<tool_name>", "args": {{...}}, "correlation_id": "c1"}}, "assistant_message": ""}}

### Вариант B — финальный ответ:
{{"type": "final", "final_message": "<полный ответ пользователю>", "citations": [], "next_suggestions": []}}

{tool_descs}

ВАЖНО:
- Отвечай ТОЛЬКО JSON. Никакого текста до или после.
- Если нужно вызвать инструмент — используй вариант A.
- Когда готов дать финальный ответ — используй вариант B.
- Не логируй chain-of-thought. Рассуждай внутренне."""


def build_conversational_prompt(system_prompt: str) -> str:
    """Build conversational prompt."""
    return f"""{system_prompt}

## Инструкции по формату ответа

Ты ОБЯЗАН отвечать ТОЛЬКО валидным JSON:
{{"type": "final", "final_message": "<твой ответ>", "citations": [], "next_suggestions": []}}

ВАЖНО:
- Отвечай ТОЛЬКО JSON. Никакого текста до или после.
- Не логируй chain-of-thought."""


def build_planner_prompt(
    system_prompt: str,
    tools: list[ToolSpec],
) -> str:
    """Build planner prompt."""
    tool_descs = _format_tools(tools)

    return f"""{system_prompt}

## Инструкции по формату ответа

Ты ОБЯЗАН вернуть JSON-план в строгом формате:
{{
  "type": "plan",
  "goal": "<цель плана>",
  "steps": [
    {{
      "id": "step1",
      "title": "<название шага>",
      "mode": "react" или "conversational",
      "tool_hints": ["<имя_инструмента>"],
      "success_criteria": ["<критерий>"],
      "max_iterations": 4
    }}
  ],
  "final_format": "<формат финального ответа>"
}}

{tool_descs}

ВАЖНО:
- Отвечай ТОЛЬКО JSON.
- Каждый шаг — отдельная подзадача.
- mode="react" если шаг требует инструментов, "conversational" если нет."""


def build_plan_step_prompt(
    system_prompt: str,
    step_title: str,
    step_context: str,
    tools: list[ToolSpec],
) -> str:
    """Build plan step prompt."""
    tool_descs = _format_tools(tools)

    return f"""{system_prompt}

## Текущий шаг плана: {step_title}

Контекст предыдущих шагов:
{step_context}

## Инструкции по формату ответа

Ты ОБЯЗАН отвечать ТОЛЬКО валидным JSON:
- Вариант A (вызов инструмента): {{"type": "tool_call", "tool": {{"name": "<tool_name>", "args": {{...}}, "correlation_id": "c1"}}, "assistant_message": ""}}
- Вариант B (результат шага): {{"type": "final", "final_message": "<результат шага>", "citations": [], "next_suggestions": []}}

{tool_descs}

ВАЖНО: Отвечай ТОЛЬКО JSON."""


def build_final_assembly_prompt(
    system_prompt: str,
    plan_goal: str,
    step_results: list[str],
    final_format: str,
) -> str:
    """Build final assembly prompt."""
    results_text = "\n\n".join(f"### Шаг {i + 1}\n{r}" for i, r in enumerate(step_results))

    return f"""{system_prompt}

## Финальная сборка

Цель: {plan_goal}

Результаты шагов:
{results_text}

Формат ответа: {final_format or "Структурированный ответ"}

## Инструкции

Ты ОБЯЗАН ответить ТОЛЬКО валидным JSON:
{{"type": "final", "final_message": "<финальный структурированный ответ>", "citations": [], "next_suggestions": []}}

Собери все результаты шагов в единый связный ответ."""


def _format_tools(tools: list[ToolSpec]) -> str:
    """Format tools."""
    if not tools:
        return "Инструменты: нет доступных."

    lines = ["## Доступные инструменты:"]
    for t in tools:
        lines.append(f"- **{t.name}**: {t.description}")
    return "\n".join(lines)



"""Team Manager module."""

from __future__ import annotations

from cognitia.orchestration.team_protocol import ResumableTeamOrchestrator, TeamOrchestrator
from cognitia.orchestration.team_types import TeamConfig, TeamMessage, TeamStatus


class TeamManager:
    """Managing a team of agents - the single point of entry for orders."""

    def __init__(self, orchestrator: TeamOrchestrator) -> None:
        self._orch = orchestrator
        self._resumable_orch: ResumableTeamOrchestrator | None = None
        if isinstance(orchestrator, ResumableTeamOrchestrator):
            self._resumable_orch = orchestrator
        self._teams: dict[str, TeamConfig] = {}
        self._paused: set[tuple[str, str]] = set()  # (team_id, agent_id)

    async def start_team(self, config: TeamConfig, task: str) -> str:
        """Run tomandu."""
        team_id = await self._orch.start(config, task)
        self._teams[team_id] = config
        return team_id

    async def stop_team(self, team_id: str) -> None:
        """Stop toomand."""
        await self._orch.stop(team_id)
        self._teams.pop(team_id, None)

    async def get_status(self, team_id: str) -> TeamStatus:
        """Get status teams."""
        return await self._orch.get_team_status(team_id)

    async def list_teams(self) -> list[str]:
        """List teams."""
        return list(self._teams.keys())

    async def send_to_agent(self, team_id: str, message: TeamMessage) -> None:
        """Send to agent."""
        await self._orch.send_message(team_id, message)

    async def pause_agent(self, team_id: str, agent_id: str) -> None:
        """Prandstop worker."""
        await self._orch.pause_agent(team_id, agent_id)
        self._paused.add((team_id, agent_id))

    async def resume_agent(self, team_id: str, agent_id: str) -> None:
        """Resume the suspended worker via typesafe contract."""
        if self._resumable_orch is None:
            return
        await self._resumable_orch.resume_agent(team_id, agent_id)
        self._paused.discard((team_id, agent_id))



"""Builtin Tools - built-in tools for agents.

Sandbox tools (bash, read, write, edit, multi_edit, ls, glob, grep)
run through SandboxProvider. Web tools (web_fetch, web_search)
run through WebProvider. Each group is created by a separate factory.

KISS: each executor <=30 lines.
"""

from __future__ import annotations

import json
import re
from collections.abc import Callable
from typing import Any

import structlog

from cognitia.runtime.types import ToolSpec
from cognitia.tools.protocols import SandboxProvider

_log = structlog.get_logger(component="web_tools")

# ---------------------------------------------------------------------------
# Alias map: SDK names -> canonical snake_case
# ---------------------------------------------------------------------------

TOOL_ALIAS_MAP: dict[str, str] = {
    "Bash": "bash",
    "Read": "read",
    "Write": "write",
    "Edit": "edit",
    "MultiEdit": "multi_edit",
    "Glob": "glob",
    "Grep": "grep",
    "LS": "ls",
    "WebFetch": "web_fetch",
    "WebSearch": "web_search",
}

# ---------------------------------------------------------------------------
# JSON Schemas
# ---------------------------------------------------------------------------

_BASH_SCHEMA: dict[str, Any] = {
    "type": "object",
    "properties": {
        "command": {"type": "string", "description": "Shell-команда для выполнения"},
    },
    "required": ["command"],
}

_READ_SCHEMA: dict[str, Any] = {
    "type": "object",
    "properties": {
        "path": {"type": "string", "description": "Относительный путь к файлу"},
    },
    "required": ["path"],
}

_WRITE_SCHEMA: dict[str, Any] = {
    "type": "object",
    "properties": {
        "path": {"type": "string", "description": "Относительный путь к файлу"},
        "content": {"type": "string", "description": "Содержимое файла"},
    },
    "required": ["path", "content"],
}

_EDIT_SCHEMA: dict[str, Any] = {
    "type": "object",
    "properties": {
        "path": {"type": "string", "description": "Путь к файлу"},
        "old_string": {"type": "string", "description": "Строка для замены"},
        "new_string": {"type": "string", "description": "Строка-замена"},
    },
    "required": ["path", "old_string", "new_string"],
}

_MULTI_EDIT_SCHEMA: dict[str, Any] = {
    "type": "object",
    "properties": {
        "path": {"type": "string", "description": "Путь к файлу"},
        "edits": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "old_string": {"type": "string"},
                    "new_string": {"type": "string"},
                },
                "required": ["old_string", "new_string"],
            },
            "description": "Массив замен [{old_string, new_string}]",
        },
    },
    "required": ["path", "edits"],
}

_LS_SCHEMA: dict[str, Any] = {
    "type": "object",
    "properties": {
        "path": {"type": "string", "description": "Путь к директории", "default": "."},
    },
}

_GLOB_SCHEMA: dict[str, Any] = {
    "type": "object",
    "properties": {
        "pattern": {"type": "string", "description": "Glob-паттерн (e.g. **/*.py)"},
    },
    "required": ["pattern"],
}

_GREP_SCHEMA: dict[str, Any] = {
    "type": "object",
    "properties": {
        "pattern": {"type": "string", "description": "Regex-паттерн для поиска"},
        "path": {"type": "string", "description": "Путь к файлу (опционально)"},
    },
    "required": ["pattern"],
}

_WEB_FETCH_SCHEMA: dict[str, Any] = {
    "type": "object",
    "properties": {
        "url": {"type": "string", "description": "URL для загрузки"},
    },
    "required": ["url"],
}

_WEB_SEARCH_SCHEMA: dict[str, Any] = {
    "type": "object",
    "properties": {
        "query": {"type": "string", "description": "Поисковый запрос"},
        "max_results": {"type": "integer", "default": 5},
    },
    "required": ["query"],
}


# ---------------------------------------------------------------------------
# Sandbox Tool Executors
# ---------------------------------------------------------------------------


def _make_json_error(message: str) -> str:
    """Create a JSON error response."""
    return json.dumps({"status": "error", "message": message})


def _create_bash_executor(sandbox: SandboxProvider) -> Callable:
    async def executor(args: dict) -> str:
        command = args.get("command")
        if not command:
            return _make_json_error("command обязателен")
        try:
            result = await sandbox.execute(command)
            return json.dumps(
                {
                    "stdout": result.stdout,
                    "stderr": result.stderr,
                    "exit_code": result.exit_code,
                    "timed_out": result.timed_out,
                }
            )
        except Exception as e:
            return _make_json_error(str(e))

    return executor


def _create_read_executor(sandbox: SandboxProvider) -> Callable:
    async def executor(args: dict) -> str:
        path = args.get("path")
        if not path:
            return _make_json_error("path обязателен")
        try:
            content = await sandbox.read_file(path)
            return json.dumps({"status": "ok", "content": content})
        except Exception as e:
            return _make_json_error(str(e))

    return executor


def _create_write_executor(sandbox: SandboxProvider) -> Callable:
    async def executor(args: dict) -> str:
        path = args.get("path")
        content = args.get("content")
        if not path or content is None:
            return _make_json_error("path и content обязательны")
        try:
            await sandbox.write_file(path, content)
            return json.dumps({"status": "ok", "path": path})
        except Exception as e:
            return _make_json_error(str(e))

    return executor


def _create_edit_executor(sandbox: SandboxProvider) -> Callable:
    async def executor(args: dict) -> str:
        path = args.get("path", "")
        old_string = args.get("old_string", "")
        new_string = args.get("new_string", "")
        if not path or not old_string:
            return _make_json_error("path и old_string обязательны")
        try:
            content = await sandbox.read_file(path)
            if old_string not in content:
                return _make_json_error(f"old_string не найден в {path}")
            # Replace only the first occurrence
            updated = content.replace(old_string, new_string, 1)
            await sandbox.write_file(path, updated)
            return json.dumps({"status": "ok", "path": path})
        except Exception as e:
            return _make_json_error(str(e))

    return executor


def _create_multi_edit_executor(sandbox: SandboxProvider) -> Callable:
    async def executor(args: dict) -> str:
        path = args.get("path", "")
        edits = args.get("edits", [])
        if not path or not edits:
            return _make_json_error("path и edits обязательны")
        try:
            content = await sandbox.read_file(path)
            for edit in edits:
                old = edit.get("old_string", "")
                new = edit.get("new_string", "")
                if old not in content:
                    return _make_json_error(f"old_string не найден: {old[:50]}")
                content = content.replace(old, new, 1)
            await sandbox.write_file(path, content)
            return json.dumps({"status": "ok", "path": path, "edits_applied": len(edits)})
        except Exception as e:
            return _make_json_error(str(e))

    return executor


def _create_ls_executor(sandbox: SandboxProvider) -> Callable:
    async def executor(args: dict) -> str:
        path = args.get("path", ".")
        try:
            entries = await sandbox.list_dir(path)
            return json.dumps({"status": "ok", "entries": entries})
        except Exception as e:
            return _make_json_error(str(e))

    return executor


def _create_glob_executor(sandbox: SandboxProvider) -> Callable:
    async def executor(args: dict) -> str:
        pattern = args.get("pattern", "")
        if not pattern:
            return _make_json_error("pattern обязателен")
        try:
            matches = await sandbox.glob_files(pattern)
            return json.dumps({"status": "ok", "matches": matches})
        except Exception as e:
            return _make_json_error(str(e))

    return executor


def _create_grep_executor(sandbox: SandboxProvider) -> Callable:
    async def executor(args: dict) -> str:
        pattern = args.get("pattern", "")
        path = args.get("path")
        if not pattern:
            return _make_json_error("pattern обязателен")
        try:
            if path:
                content = await sandbox.read_file(path)
                matches = re.findall(pattern, content)
                return json.dumps({"status": "ok", "matches": matches, "path": path})
            # Without a path, search across all workspace files (basic implementation)
            return json.dumps({"status": "ok", "matches": [], "note": "path рекомендуется"})
        except Exception as e:
            return _make_json_error(str(e))

    return executor


# ---------------------------------------------------------------------------
# Factory: sandbox tools
# ---------------------------------------------------------------------------


def create_sandbox_tools(
    sandbox: SandboxProvider,
) -> tuple[dict[str, ToolSpec], dict[str, Callable]]:
    """Create sandbox tools.

    Returns:
        Tuple: (specs dict, executors dict).
    """
    tools: list[tuple[str, str, dict, Callable]] = [
        ("bash", "Выполнить shell-команду в sandbox", _BASH_SCHEMA, _create_bash_executor(sandbox)),
        ("read", "Прочитать файл из workspace", _READ_SCHEMA, _create_read_executor(sandbox)),
        ("write", "Записать файл в workspace", _WRITE_SCHEMA, _create_write_executor(sandbox)),
        (
            "edit",
            "Заменить подстроку в файле (str_replace)",
            _EDIT_SCHEMA,
            _create_edit_executor(sandbox),
        ),
        (
            "multi_edit",
            "Несколько замен в одном файле",
            _MULTI_EDIT_SCHEMA,
            _create_multi_edit_executor(sandbox),
        ),
        ("ls", "Список файлов в директории workspace", _LS_SCHEMA, _create_ls_executor(sandbox)),
        ("glob", "Поиск файлов по glob-паттерну", _GLOB_SCHEMA, _create_glob_executor(sandbox)),
        ("grep", "Поиск текста по regex в файлах", _GREP_SCHEMA, _create_grep_executor(sandbox)),
    ]

    specs: dict[str, ToolSpec] = {}
    executors: dict[str, Callable] = {}

    for name, desc, schema, executor in tools:
        specs[name] = ToolSpec(name=name, description=desc, parameters=schema)
        executors[name] = executor

    return specs, executors


# ---------------------------------------------------------------------------
# Factory: web tools
# ---------------------------------------------------------------------------


def create_web_tools(
    web_provider: Any | None,
) -> tuple[dict[str, ToolSpec], dict[str, Callable]]:
    """Create web tools.

    Args:
        web_provider: WebProvider or None. If None, tools are not created.

    Returns:
        Tuple: (specs dict, executors dict). Empty if no provider is configured.
    """
    if web_provider is None:
        return {}, {}

    specs: dict[str, ToolSpec] = {}
    executors: dict[str, Callable] = {}

    async def fetch_executor(args: dict) -> str:
        url = args.get("url", "")
        if not url or not url.strip():
            return _make_json_error("url обязателен")
        try:
            content = await web_provider.fetch(url.strip())
            return json.dumps({"status": "ok", "content": content})
        except Exception as e:
            _log.warning("web_fetch_failed", url=url[:200], error=str(e))
            return _make_json_error(str(e))

    async def search_executor(args: dict) -> str:
        query = args.get("query", "")
        max_results = args.get("max_results", 5)
        if not query or not query.strip():
            return _make_json_error("query обязателен")
        try:
            results = await web_provider.search(query.strip(), max_results)
            return json.dumps(
                {
                    "status": "ok",
                    "result_count": len(results),
                    "results": [
                        {"title": r.title, "url": r.url, "snippet": r.snippet} for r in results
                    ],
                }
            )
        except Exception as e:
            _log.warning("web_search_failed", query=query[:100], error=str(e))
            return json.dumps({"status": "error", "message": str(e)})

    specs["web_fetch"] = ToolSpec(
        name="web_fetch", description="Получить содержимое URL", parameters=_WEB_FETCH_SCHEMA
    )
    specs["web_search"] = ToolSpec(
        name="web_search", description="Поиск в интернете", parameters=_WEB_SEARCH_SCHEMA
    )
    executors["web_fetch"] = fetch_executor
    executors["web_search"] = search_executor

    return specs, executors



"""ToolSelector - smart tool selection under a context budget.

Problem: 40+ tools = 5000-7000 tokens just for schema.
Solution: priority groups + configurable budget.

Everything is configured through ToolBudgetConfig:
- max_tools: overall limit
- group_priority: priority order (overridable)
- group_limits: per-group limit (optional)
"""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import IntEnum

from cognitia.runtime.types import ToolSpec


class ToolGroup(IntEnum):
    """Priority tool groups.

    Lower value = higher priority.
    The default order can be overridden via ToolBudgetConfig.group_priority.
    """

    ALWAYS = 0  # thinking, todo - always included
    MCP = 1  # current-role MCP tools - business logic
    MEMORY = 2  # memory_* tools
    PLANNING = 3  # plan_* tools
    SANDBOX = 4  # bash, read, write, edit, ...
    WEB = 5  # web_fetch, web_search


@dataclass(frozen=True)
class ToolBudgetConfig:
    """Tool budget configuration.

    All settings are moved out of hardcode - library users
    control the budget through config.
    """

    max_tools: int = 30
    group_priority: list[ToolGroup] = field(
        default_factory=lambda: [
            ToolGroup.ALWAYS,
            ToolGroup.MCP,
            ToolGroup.MEMORY,
            ToolGroup.PLANNING,
            ToolGroup.SANDBOX,
            ToolGroup.WEB,
        ]
    )
    group_limits: dict[ToolGroup, int] = field(default_factory=dict)


class ToolSelector:
    """Selects tools by priority and budget.

    Fills the budget from top to bottom by groups from config.group_priority.
    If group_limits is set, limits each group.
    """

    def __init__(self, config: ToolBudgetConfig | None = None, *, max_tools: int = 30) -> None:
        if config is not None:
            self._config = config
        else:
            self._config = ToolBudgetConfig(max_tools=max_tools)
        self._groups: dict[ToolGroup, list[ToolSpec]] = {}

    def add_group(self, group: ToolGroup, tools: list[ToolSpec]) -> None:
        """Add a tool group."""
        self._groups[group] = tools

    def select(self) -> list[ToolSpec]:
        """Select tools within the budget.

        Returns:
            List of ToolSpec objects sorted by priority from config.
        """
        result: list[ToolSpec] = []
        remaining = self._config.max_tools

        # Traverse groups in the order specified by the config
        for group in self._config.group_priority:
            tools = self._groups.get(group, [])
            if remaining <= 0 or not tools:
                continue

            # Per-group limit (if set)
            group_limit = self._config.group_limits.get(group, remaining)
            take = min(len(tools), remaining, group_limit)

            result.extend(tools[:take])
            remaining -= take

        return result



"""Plan Store module."""

from __future__ import annotations

from cognitia.orchestration.types import Plan, PlanStep


class InMemoryPlanStore:
    """In-memory is real from the PlanStore with multi-tenant from the area.

  Namespace is set oninlanded via set_namespace(user_id, topic_id).
  save() aintomatandchestoandprandinyazyinaplan to the current namespace.
  list_plans() filters by namespace.
  """

    def __init__(self) -> None:
        self._plans: dict[str, Plan] = {}
        self._ownership: dict[str, tuple[str, str]] = {}  # plan_id → (user_id, topic_id)
        self._ns_user: str = ""
        self._ns_topic: str = ""

    def set_namespace(self, user_id: str, topic_id: str) -> None:
        """Set up the namespace for save/list."""
        self._ns_user = user_id
        self._ns_topic = topic_id

    async def save(self, plan: Plan) -> None:
        """Save or update plan. Prandin speaks to the current namespace."""
        self._plans[plan.id] = plan
        if self._ns_user or self._ns_topic:
            self._ownership[plan.id] = (self._ns_user, self._ns_topic)

    async def load(self, plan_id: str) -> Plan | None:
        """Load."""
        return self._plans.get(plan_id)

    async def list_plans(self, user_id: str, topic_id: str) -> list[Plan]:
        """List plans."""
        if not user_id and not topic_id:
            return list(self._plans.values())
        result: list[Plan] = []
        for plan_id, plan in self._plans.items():
            owner = self._ownership.get(plan_id, ("", ""))
            if (not user_id or owner[0] == user_id) and (not topic_id or owner[1] == topic_id):
                result.append(plan)
        return result

    async def update_step(self, plan_id: str, step: PlanStep) -> None:
        """Update step."""
        plan = self._plans.get(plan_id)
        if plan is None:
            return
        try:
            updated = plan.update_step(step)
            self._plans[plan_id] = updated
        except ValueError:
            pass



"""Agent-as-tool: expose a runtime as a callable tool for other agents.

Provides two functions:
- execute_agent_tool: run an agent runtime and collect the final result
- create_agent_tool_spec: build a ToolSpec for an agent-as-tool
"""

from __future__ import annotations

import asyncio
from typing import Any, Callable

from cognitia.multi_agent.types import AgentToolResult
from cognitia.runtime.types import ToolSpec, Message


async def execute_agent_tool(
    run_fn: Callable[..., Any],
    query: str,
    system_prompt: str = "You are a helpful assistant.",
    timeout_seconds: float = 60.0,
) -> AgentToolResult:
    """Execute an agent runtime as a tool, collecting final result.

    Args:
        run_fn: Async generator function with signature
            (messages, system_prompt, active_tools) -> AsyncIterator[RuntimeEvent].
        query: User query to send to the sub-agent.
        system_prompt: System prompt for the sub-agent.
        timeout_seconds: Maximum execution time before timeout.

    Returns:
        AgentToolResult with success/failure and output text.
    """
    messages = [Message(role="user", content=query)]

    async def _run() -> str:
        final_text = ""
        final_seen = False
        async for event in run_fn(
            messages=messages, system_prompt=system_prompt, active_tools=[],
        ):
            if event.is_error:
                error_message = event.data.get("message") or "Sub-agent runtime failed"
                raise RuntimeError(error_message)
            if event.is_final:
                final_text = event.text or ""
                final_seen = True
                break
        if not final_seen:
            raise RuntimeError("Sub-agent runtime ended without a final event")
        return final_text

    try:
        final_text = await asyncio.wait_for(_run(), timeout=timeout_seconds)
        return AgentToolResult(success=True, output=final_text)
    except asyncio.TimeoutError:
        return AgentToolResult(success=False, output="", error="Timeout exceeded")
    except Exception as e:
        return AgentToolResult(success=False, output="", error=str(e))


def create_agent_tool_spec(name: str, description: str) -> ToolSpec:
    """Create a ToolSpec for an agent-as-tool.

    Args:
        name: Tool name visible to the orchestrating agent.
        description: Tool description for the LLM.

    Returns:
        ToolSpec with a single required 'query' string parameter.
    """
    return ToolSpec(
        name=name,
        description=description,
        parameters={
            "type": "object",
            "properties": {
                "query": {
                    "type": "string",
                    "description": "The query to send to the sub-agent",
                },
            },
            "required": ["query"],
        },
        is_local=True,
    )



"""SearXNG search provider - self-hosted metasearch engine.

Requires a SearXNG instance URL (SEARXNG_URL).
Dependency: httpx (already included in cognitia[web]).
"""

from __future__ import annotations

import httpx
import structlog

from cognitia.tools.web_protocols import SearchResult

_log = structlog.get_logger(component="web_search.searxng")


class SearXNGSearchProvider:
    """Search via SearXNG (self-hosted, no limits).

    Connects to a user-provided SearXNG instance by URL.
    """

    def __init__(self, base_url: str, timeout: int = 15) -> None:
        if not base_url:
            raise ValueError("SEARXNG_URL обязателен для SearXNGSearchProvider")
        self._base_url = base_url.rstrip("/")
        self._timeout = timeout

    async def search(self, query: str, max_results: int = 5) -> list[SearchResult]:
        """Search via the SearXNG JSON API.

        Args:
            query: Search query. Empty/whitespace -> empty list.
            max_results: Maximum number of results.

        Returns:
            List of SearchResult. Empty on connection error.
        """
        if not query or not query.strip():
            return []

        params: dict[str, str] = {
            "q": query,
            "format": "json",
            "pageno": "1",
        }

        try:
            async with httpx.AsyncClient(timeout=self._timeout) as client:
                response = await client.get(
                    f"{self._base_url}/search",
                    params=params,
                )
                response.raise_for_status()
                data = response.json()
        except (httpx.HTTPError, ValueError, OSError) as exc:
            _log.warning("searxng_search_failed", query=query[:100], error=str(exc))
            return []

        results = data.get("results", [])[:max_results]
        return [
            SearchResult(
                title=r.get("title", ""),
                url=r.get("url", ""),
                snippet=r.get("content", ""),
            )
            for r in results
        ]



"""SessionRehydrator - restore session state from Postgres (architecture section 8.4).

Contract: build_rehydration_payload(ctx: TurnContext) -> Mapping
Restores: role_id, active_skill_ids, summary, last_N messages, goals, phase, prompt_hash.

ISP/DIP: depends on small Protocols (<=5 methods each), not a monolithic provider.
"""

from __future__ import annotations

from typing import Any

from cognitia.protocols import (
    GoalStore,
    MessageStore,
    PhaseStore,
    SessionStateStore,
    SummaryStore,
)
from cognitia.types import TurnContext


class DefaultSessionRehydrator:
    """Postgres rehydration implementation.

    Order (from section 8.4):
    1. Session state (role_id, skills) from the topics table
    2. Rolling summary
    3. Last N messages (current topic)
    4. Active goal + phase_state

    Conflict resolution: summary vs last_N -> last_N wins.

    ISP: depends on 5 small protocols rather than a monolithic MemoryProvider.
    """

    def __init__(
        self,
        messages: MessageStore,
        summaries: SummaryStore,
        goals: GoalStore,
        sessions: SessionStateStore,
        phases: PhaseStore,
        last_n_messages: int = 10,
    ) -> None:
        self._messages = messages
        self._summaries = summaries
        self._goals = goals
        self._sessions = sessions
        self._phases = phases
        self._last_n = last_n_messages

    async def build_rehydration_payload(self, ctx: TurnContext) -> dict[str, Any]:
        """Build the full rehydration payload (§8.4).

        Order:
        1. Session state (role_id, skills) from the topics table
        2. Rolling summary
        3. Last N messages (current topic)
        4. Active goal + phase_state
        """
        # 1. Session state from the database
        session_state = await self._sessions.get_session_state(ctx.user_id, ctx.topic_id)

        if session_state:
            role_id = session_state["role_id"]
            active_skill_ids = session_state.get("active_skill_ids", [])
            prompt_hash = session_state.get("prompt_hash", "")
        else:
            # No database data available - fall back to ctx
            role_id = ctx.role_id
            active_skill_ids = list(ctx.active_skill_ids) if ctx.active_skill_ids else []
            prompt_hash = ""

        # 2. Summary
        summary = await self._summaries.get_summary(ctx.user_id, ctx.topic_id)

        # 3. Last N messages
        last_messages = await self._messages.get_messages(
            ctx.user_id, ctx.topic_id, limit=self._last_n
        )

        # 4. Active goal
        goal = await self._goals.get_active_goal(ctx.user_id, ctx.topic_id)

        # 5. Phase state
        phase_state = await self._phases.get_phase_state(ctx.user_id)

        return {
            "role_id": role_id,
            "active_skill_ids": active_skill_ids,
            "prompt_hash": prompt_hash,
            "summary": summary,
            "last_messages": last_messages,
            "goal": goal,
            "phase_state": phase_state,
        }



"""DuckDuckGo search provider - metasearch across 9 engines, no API keys.

Optional dependency: ddgs (pip install cognitia[web-duckduckgo]).
Engines: Bing, Google, Brave, DuckDuckGo, Yandex, Yahoo, Mojeek, Wikipedia, Grokipedia.
If the dependency is not installed, gracefully fall back to an empty list.
"""

from __future__ import annotations

import asyncio

import structlog

from cognitia.tools.web_protocols import SearchResult

try:
    from ddgs import DDGS  # type: ignore[import-not-found]
except ImportError:
    DDGS = None  # type: ignore[assignment,misc]

_log = structlog.get_logger(component="web_search.duckduckgo")


class DuckDuckGoSearchProvider:
    """Metasearch via ddgs (9 engines, no API keys).

    Uses the ddgs library (formerly duckduckgo-search).
    Requests are executed synchronously via run_in_executor.
    """

    def __init__(self, timeout: int = 15) -> None:
        self._timeout = timeout

    async def search(self, query: str, max_results: int = 5) -> list[SearchResult]:
        """Search via ddgs.

        Args:
            query: Search query. Empty/whitespace -> empty list.
            max_results: Maximum number of results.

        Returns:
            List of SearchResult. Empty if ddgs is unavailable or an error occurs.
        """
        if DDGS is None:
            return []
        if not query or not query.strip():
            return []

        def _sync_search() -> list[dict]:
            try:
                return list(DDGS().text(query, max_results=max_results, timeout=self._timeout))
            except Exception as exc:
                _log.warning("ddgs_search_failed", query=query[:100], error=str(exc))
                return []

        loop = asyncio.get_running_loop()
        raw_results = await loop.run_in_executor(None, _sync_search)

        return [
            SearchResult(
                title=r.get("title", ""),
                url=r.get("href", ""),
                snippet=r.get("body", ""),
            )
            for r in raw_results
        ]



"""Types for the Todo capability.

TodoItem - checklist item (id, content, status).
TodoConfig - todo configuration (enabled, backend, limits).
"""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Literal


@dataclass(frozen=True)
class TodoItem:
    """Checklist item.

    Statuses: pending, in_progress, completed, cancelled.
    """

    id: str
    content: str
    status: Literal["pending", "in_progress", "completed", "cancelled"]
    created_at: datetime
    updated_at: datetime

    def to_dict(self) -> dict[str, Any]:
        """Serialize to a dict."""
        return {
            "id": self.id,
            "content": self.content,
            "status": self.status,
            "created_at": self.created_at.isoformat(),
            "updated_at": self.updated_at.isoformat(),
        }


@dataclass(frozen=True)
class TodoConfig:
    """Todo capability configuration.

    Can be enabled or disabled independently of sandbox and memory bank.
    """

    enabled: bool = False
    backend: Literal["memory", "filesystem", "database"] = "memory"
    root_path: Path | None = None
    max_todos: int = 100
    auto_cleanup_completed: bool = False



"""Cost budget tracking for runtime execution.

Provides:
- ModelPricing: per-model token pricing (USD per 1M tokens)
- CostBudget: budget limits configuration
- CostTracker: accumulates usage and checks budget
- load_pricing(): loads pricing.JSON via importlib.resources
"""

from __future__ import annotations

import importlib.resources
import json
from dataclasses import dataclass
from typing import Literal

BudgetStatus = Literal["ok", "warning", "exceeded"]


@dataclass(frozen=True)
class ModelPricing:
    """Per-model token pricing in USD per 1M tokens."""

    input_per_1m: float
    output_per_1m: float


@dataclass(frozen=True)
class CostBudget:
    """Budget limits for cost tracking.

  Attributes:
    max_cost_usd: Maximum total cost in USD. None = no limit.
    max_total_tokens: Maximum total tokens (input + output). None = no limit.
    action_on_exceed: What to do when budget is exceeded.
      "error" = emit budget_exceeded error event.
      "warn" = report warning status but continue.
  """

    max_cost_usd: float | None = None
    max_total_tokens: int | None = None
    action_on_exceed: Literal["error", "warn"] = "error"


class CostTracker:
    """Accumulates token usage and checks budget limits.

  Thread-safe for single-threaded async usage (no locks needed).
  """

    def __init__(self, budget: CostBudget, pricing: dict[str, ModelPricing]) -> None:
        self._budget = budget
        self._pricing = pricing
        self._total_input_tokens: int = 0
        self._total_output_tokens: int = 0
        self._total_cost: float = 0.0

    def record(self, model: str, input_tokens: int, output_tokens: int) -> None:
        """Record token usage for a single LLM call."""
        pricing = self._pricing.get(model) or self._pricing.get("_default")
        if pricing is None:
            return

        cost = (
            pricing.input_per_1m * input_tokens / 1_000_000
            + pricing.output_per_1m * output_tokens / 1_000_000
        )
        self._total_input_tokens += input_tokens
        self._total_output_tokens += output_tokens
        self._total_cost += cost

    @property
    def total_cost_usd(self) -> float:
        """Total accumulated cost in USD."""
        return self._total_cost

    @property
    def total_tokens(self) -> int:
        """Total accumulated tokens (input + output)."""
        return self._total_input_tokens + self._total_output_tokens

    def check_budget(self) -> BudgetStatus:
        """Check whether budget limits have been exceeded.

    Returns:
      "ok" - wiThin limits (or no limits set).
      "exceeded" - over limit with action_on_exceed="error".
      "warning" - over limit with action_on_exceed="warn".
    """
        exceeded = False
        if self._budget.max_cost_usd is not None and self._total_cost > self._budget.max_cost_usd:
            exceeded = True
        if (
            self._budget.max_total_tokens is not None
            and self.total_tokens > self._budget.max_total_tokens
        ):
            exceeded = True

        if not exceeded:
            return "ok"

        if self._budget.action_on_exceed == "warn":
            return "warning"
        return "exceeded"

    def reset(self) -> None:
        """Reset all accumulated usage to zero."""
        self._total_input_tokens = 0
        self._total_output_tokens = 0
        self._total_cost = 0.0


def load_pricing() -> dict[str, ModelPricing]:
    """Load model pricing from bundled pricing.JSON.

  Uses importlib.resources for reliable package-relative loading.
  """
    ref = importlib.resources.files("cognitia.runtime").joinpath("pricing.json")
    raw = ref.read_text(encoding="utf-8")
    data: dict[str, dict[str, float]] = json.loads(raw)
    return {
        model: ModelPricing(input_per_1m=info["input_per_1m"], output_per_1m=info["output_per_1m"])
        for model, info in data.items()
    }



"""Types for agent sandbox isolation.

SandboxConfig is the sandbox configuration (root_path, user/topic isolation, limits).
ExecutionResult is the result of command execution.
SandboxViolation is raised on isolation violations.
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path


@dataclass(frozen=True)
class SandboxConfig:
    """Sandbox isolation configuration.

    The sandbox isolates filesystem access and command execution
    by user_id / topic_id. Each agent works in its own workspace:
    {root_path}/{user_id}/{topic_id}/workspace/
    """

    root_path: str
    user_id: str
    topic_id: str
    max_file_size_bytes: int = 10 * 1024 * 1024  # 10 MB
    timeout_seconds: int = 30
    allowed_extensions: frozenset[str] | None = None
    denied_commands: frozenset[str] | None = None

    @property
    def workspace_path(self) -> str:
        """Absolute path to the agent workspace."""
        return str(Path(self.root_path) / self.user_id / self.topic_id / "workspace")


@dataclass(frozen=True)
class ExecutionResult:
    """Result of executing a command in the sandbox.

    Contains stdout, stderr, exit_code, and a timeout flag.
    """

    stdout: str
    stderr: str
    exit_code: int
    timed_out: bool


class SandboxViolation(Exception):
    """Sandbox isolation violation.

    Raised when path traversal is attempted, limits are exceeded,
    or denied commands are executed.
    """

    def __init__(self, message: str, *, path: str | None = None) -> None:
        super().__init__(message)
        self.path = path



"""HttpxWebProvider - WebProvider implementation via httpx.

fetch: GET URL -> trafilatura/regex -> text (always works).
search: delegates to a pluggable WebSearchProvider (DIP).
Optional dependency: httpx, trafilatura.
"""

from __future__ import annotations

import re

import structlog

from cognitia.tools.web_protocols import SearchResult, WebFetchProvider, WebSearchProvider

try:
    import trafilatura
except ImportError:
    trafilatura = None  # type: ignore[assignment]

_log = structlog.get_logger(component="web_httpx")


def _extract_text(html: str) -> str:
    """Extract text from HTML: trafilatura -> improved regex fallback.

    Trafilatura (if installed) extracts the main page content,
    discarding navigation, footers, and ads.
    Without trafilatura, regex removes script/style/tags.
    """
    if trafilatura is not None:
        text = trafilatura.extract(html, include_links=True) or ""
        if text:
            return text[:50000]

    # Improved regex fallback: remove script, style, then tags
    text = re.sub(r"<script[^>]*>.*?</script>", "", html, flags=re.DOTALL)
    text = re.sub(r"<style[^>]*>.*?</style>", "", text, flags=re.DOTALL)
    text = re.sub(r"<[^>]+>", " ", text)
    text = re.sub(r"\s+", " ", text).strip()
    return text[:50000]


class HttpxWebProvider:
    """WebProvider via httpx (async HTTP client).

    fetch: GET -> trafilatura/regex -> text. Or delegate to fetch_provider (Jina/Crawl4AI).
    search: delegates to search_provider (DuckDuckGo, Tavily, SearXNG, Brave).
    """

    def __init__(
        self,
        timeout: int = 30,
        search_provider: WebSearchProvider | None = None,
        fetch_provider: WebFetchProvider | None = None,
    ) -> None:
        self._timeout = timeout
        self._search_provider = search_provider
        self._fetch_provider = fetch_provider

    async def fetch(self, url: str) -> str:
        """Load a URL and return text content.

        If fetch_provider (Jina/Crawl4AI) is set, delegate to it.
        Otherwise, use httpx GET + trafilatura/regex.
        """
        if self._fetch_provider is not None:
            return await self._fetch_provider.fetch(url)

        try:
            import httpx
        except ImportError:
            return "httpx не установлен. pip install cognitia[web]"

        try:
            async with httpx.AsyncClient(timeout=self._timeout, follow_redirects=True) as client:
                response = await client.get(url)
                response.raise_for_status()
                html = response.text
                return _extract_text(html)
        except Exception as exc:
            _log.warning("httpx_fetch_failed", url=url[:200], error=str(exc))
            return ""

    async def search(self, query: str, max_results: int = 5) -> list[SearchResult]:
        """Search the internet via a pluggable provider.

        Delegates to search_provider when it is set.
        Returns an empty list when no provider is configured.
        """
        if self._search_provider is None:
            return []
        return await self._search_provider.search(query, max_results)



"""Context module - system_prompt assembly with budgeting."""

from cognitia.context.budget import ContextBudget, estimate_tokens, truncate_to_budget
from cognitia.context.builder import (
    BuiltContext,
    ContextInput,
    DefaultContextBuilder,
    compute_prompt_hash,
)

__all__ = [
    "BuiltContext",
    "ContextBudget",
    "ContextInput",
    "DefaultContextBuilder",
    "compute_prompt_hash",
    "estimate_tokens",
    "truncate_to_budget",
]



"""Planner strategy -- plan JSON -> step execution -> final assembly."""

from __future__ import annotations

from collections.abc import AsyncIterator, Callable
from typing import Any

from cognitia.runtime.structured_output import append_structured_output_instruction
from cognitia.runtime.thin.conversational import run_conversational
from cognitia.runtime.thin.errors import ThinLlmError
from cognitia.runtime.thin.executor import ToolExecutor
from cognitia.runtime.thin.finalization import CheckpointFn, finalize_with_validation
from cognitia.runtime.thin.helpers import _messages_to_lm
from cognitia.runtime.thin.parsers import parse_envelope, parse_plan
from cognitia.runtime.thin.prompts import (
    build_final_assembly_prompt,
    build_plan_step_prompt,
    build_planner_prompt,
)
from cognitia.runtime.thin.react_strategy import run_react
from cognitia.runtime.types import (
    Message,
    RuntimeConfig,
    RuntimeErrorData,
    RuntimeEvent,
    ToolSpec,
)


async def run_planner(
    llm_call: Callable[..., Any],
    executor: ToolExecutor,
    messages: list[Message],
    system_prompt: str,
    tools: list[ToolSpec],
    config: RuntimeConfig,
    start_time: float,
    checkpoint: CheckpointFn | None = None,
) -> AsyncIterator[RuntimeEvent]:
    """Planner-lite: plan -> step execution -> final assembly."""
    # Step 1: get plan from LLM
    prompt = build_planner_prompt(system_prompt, tools)
    lm_messages = _messages_to_lm(messages)

    try:
        checkpoint_event = await _run_checkpoint(checkpoint)
        if checkpoint_event is not None:
            yield checkpoint_event
            return
        raw = await llm_call(lm_messages, prompt)
    except ThinLlmError as exc:
        yield RuntimeEvent.error(exc.error)
        return
    checkpoint_event = await _run_checkpoint(checkpoint)
    if checkpoint_event is not None:
        yield checkpoint_event
        return
    plan = parse_plan(raw)

    if plan is None:
        # Retry
        try:
            checkpoint_event = await _run_checkpoint(checkpoint)
            if checkpoint_event is not None:
                yield checkpoint_event
                return
            raw = await llm_call(lm_messages, prompt)
        except ThinLlmError as exc:
            yield RuntimeEvent.error(exc.error)
            return
        checkpoint_event = await _run_checkpoint(checkpoint)
        if checkpoint_event is not None:
            yield checkpoint_event
            return
        plan = parse_plan(raw)

    if plan is None:
        yield RuntimeEvent.error(
            RuntimeErrorData(
                kind="bad_model_output",
                message="LLM did not return a valid plan after 2 attempts",
                recoverable=False,
            )
        )
        return

    yield RuntimeEvent.status(f"План: {plan.goal} ({len(plan.steps)} шагов)")
    steps_preview = " -> ".join(
        f"{idx}. {step.title} [{step.mode}]" for idx, step in enumerate(plan.steps, start=1)
    )
    if steps_preview:
        yield RuntimeEvent.status(f"Следующие шаги: {steps_preview}")


    step_results: list[str] = []
    new_messages: list[Message] = []
    total_tool_calls = 0

    for idx, step in enumerate(plan.steps, start=1):
        yield RuntimeEvent.status(
            f"Шаг {idx}/{len(plan.steps)}: {step.title} (режим: {step.mode})"
        )

        step_context = "\n".join(step_results) if step_results else "Нет предыдущих шагов."


        step_config = RuntimeConfig(
            runtime_name="thin",
            max_iterations=step.max_iterations,
            max_tool_calls=config.max_tool_calls - total_tool_calls,
            max_model_retries=config.max_model_retries,
            model=config.model,
        )

        step_text = ""

        if step.mode == "react":
            async for event in run_react(
                llm_call,
                executor,
                messages,
                system_prompt=build_plan_step_prompt(
                    system_prompt,
                    step.title,
                    step_context,
                    tools,
                ),
                tools=tools,
                config=step_config,
                start_time=start_time,
                checkpoint=checkpoint,
            ):

                if event.type in (
                    "tool_call_started",
                    "tool_call_finished",
                    "status",
                    "assistant_delta",
                ):
                    yield event
                elif event.type == "final":
                    step_text = event.data.get("text", "")
                    total_tool_calls += event.data.get("metrics", {}).get(
                        "tool_calls_count", 0
                    )
                elif event.type == "error":
                    yield event
                    return
        else:
            # conversational sub-step
            async for event in run_conversational(
                llm_call,
                messages,
                system_prompt=build_plan_step_prompt(
                    system_prompt,
                    step.title,
                    step_context,
                    [],
                ),
                config=step_config,
                start_time=start_time,
                checkpoint=checkpoint,
            ):
                if event.type == "assistant_delta":
                    yield event
                elif event.type == "final":
                    step_text = event.data.get("text", "")
                elif event.type == "error":
                    yield event
                    return

        step_results.append(step_text)


    assembly_prompt = build_final_assembly_prompt(
        append_structured_output_instruction(
            system_prompt,
            config.output_format,
            final_response_field="final_message",
        ),
        plan.goal,
        step_results,
        plan.final_format,
    )
    try:
        checkpoint_event = await _run_checkpoint(checkpoint)
        if checkpoint_event is not None:
            yield checkpoint_event
            return
        raw = await llm_call(lm_messages, assembly_prompt)
    except ThinLlmError as exc:
        yield RuntimeEvent.error(exc.error)
        return
    checkpoint_event = await _run_checkpoint(checkpoint)
    if checkpoint_event is not None:
        yield checkpoint_event
        return
    envelope = parse_envelope(raw)

    if envelope and envelope.type == "final" and envelope.final_message:
        final_text = envelope.final_message
    else:
        final_text = raw  # fallback

    checkpoint_event = await _run_checkpoint(checkpoint)
    if checkpoint_event is not None:
        yield checkpoint_event
        return
    yield RuntimeEvent.assistant_delta(final_text)
    async for event in finalize_with_validation(
        final_text,
        config,
        lm_messages,
        assembly_prompt,
        llm_call,
        start_time,
        iterations=len(plan.steps) + 2,
        tool_calls=total_tool_calls,
        new_messages_prefix=new_messages,
        checkpoint=checkpoint,
    ):
        yield event


async def _run_checkpoint(checkpoint: CheckpointFn | None) -> RuntimeEvent | None:
    if checkpoint is None:
        return None
    return await checkpoint()



"""PostgresMemoryProvider - MemoryProvider implementation on asyncpg + SQLAlchemy."""

from __future__ import annotations

import json
from collections.abc import AsyncIterator, Sequence
from contextlib import asynccontextmanager
from typing import Any

from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker

from cognitia.memory.types import GoalState, MemoryMessage, PhaseState, ToolEvent, UserProfile

# Subquery for resolving the internal user id by external_id (DRY)
_USER_ID_SUB = "(SELECT id FROM users WHERE external_id = :user_id)"
_POSTGRES_EXISTING_SOURCE_PRIORITY = (
    "CASE facts.source "
    "WHEN 'user' THEN 3 "
    "WHEN 'ai_inferred' THEN 2 "
    "WHEN 'mcp' THEN 1 "
    "ELSE 0 END"
)
_POSTGRES_INCOMING_SOURCE_PRIORITY = (
    "CASE EXCLUDED.source "
    "WHEN 'user' THEN 3 "
    "WHEN 'ai_inferred' THEN 2 "
    "WHEN 'mcp' THEN 1 "
    "ELSE 0 END"
)


class PostgresMemoryProvider:
    """Memory provider backed by Postgres (SQLAlchemy async).

    Implements the protocols: MessageStore, FactStore, SummaryStore,
    GoalStore, SessionStateStore, UserStore.
    """

    def __init__(self, session_factory: async_sessionmaker[AsyncSession]) -> None:
        self._sf = session_factory

    # --- Utilities (DRY) ---

    @asynccontextmanager
    async def _session(self, *, commit: bool = False) -> AsyncIterator[AsyncSession]:
        """Context manager for a DB session - removes async with duplication."""
        async with self._sf() as session:
            yield session
            if commit:
                await session.commit()

    # --- Messages (MessageStore) ---

    async def save_message(
        self,
        user_id: str,
        topic_id: str,
        role: str,
        content: str,
        tool_calls: list[dict[str, Any]] | None = None,
    ) -> None:
        """Save a message to the topic history."""
        async with self._session(commit=True) as session:
            await session.execute(
                text(
                    f"""
                    INSERT INTO messages (user_id, topic_id, role, content, tool_calls)
                    VALUES ({_USER_ID_SUB}, :topic_id, :role, :content, CAST(:tool_calls AS jsonb))
                """
                ),
                {
                    "user_id": user_id,
                    "topic_id": topic_id,
                    "role": role,
                    "content": content,
                    "tool_calls": _json_or_none(tool_calls),
                },
            )

    async def get_messages(
        self, user_id: str, topic_id: str, limit: int = 10
    ) -> list[MemoryMessage]:
        """Get the last N topic messages (ASC by time)."""
        async with self._session() as session:
            result = await session.execute(
                text(
                    f"""
                    SELECT role, content, tool_calls
                    FROM messages
                    WHERE user_id = {_USER_ID_SUB} AND topic_id = :topic_id
                    ORDER BY created_at DESC
                    LIMIT :limit
                """
                ),
                {"user_id": user_id, "topic_id": topic_id, "limit": limit},
            )
            rows = result.fetchall()
        # Reverse - from oldest to newest
        return [
            MemoryMessage(role=r.role, content=r.content, tool_calls=r.tool_calls)
            for r in reversed(rows)
        ]

    async def count_messages(self, user_id: str, topic_id: str) -> int:
        """Number of messages in the topic."""
        async with self._session() as session:
            result = await session.execute(
                text(
                    f"""
                    SELECT COUNT(*) as cnt FROM messages
                    WHERE user_id = {_USER_ID_SUB} AND topic_id = :topic_id
                """
                ),
                {"user_id": user_id, "topic_id": topic_id},
            )
            row = result.fetchone()
            return row.cnt if row else 0

    async def delete_messages_before(self, user_id: str, topic_id: str, keep_last: int = 10) -> int:
        """Delete old messages, keeping the last keep_last."""
        async with self._session(commit=True) as session:
            result = await session.execute(
                text(
                    f"""
                    DELETE FROM messages
                    WHERE id IN (
                        SELECT id FROM messages
                        WHERE user_id = {_USER_ID_SUB} AND topic_id = :topic_id
                        ORDER BY created_at ASC
                        LIMIT (
                            SELECT GREATEST(COUNT(*) - :keep_last, 0)
                            FROM messages
                            WHERE user_id = {_USER_ID_SUB} AND topic_id = :topic_id
                        )
                    )
                """
                ),
                {"user_id": user_id, "topic_id": topic_id, "keep_last": keep_last},
            )
            return int(result.rowcount)  # type: ignore[attr-defined]

    # --- Facts (FactStore) ---

    async def upsert_fact(
        self,
        user_id: str,
        key: str,
        value: Any,
        topic_id: str | None = None,
        source: str = "user",
    ) -> None:
        """Save/update a fact. Priority: user > ai_inferred > mcp.

        Uses partial unique indexes (migration 005):
        - topic_id IS NOT NULL -> ON CONFLICT on uq_facts_user_topic_key
        - topic_id IS NULL -> ON CONFLICT on uq_facts_user_global_key
        """
        json_value = json.dumps(value)
        async with self._session(commit=True) as session:
            if topic_id is not None:
                # Topic-scoped fact
                await session.execute(
                    text(
                        f"""
                        INSERT INTO facts (user_id, topic_id, key, value, source)
                        VALUES ({_USER_ID_SUB}, :topic_id, :key, CAST(:value AS jsonb), :source)
                        ON CONFLICT (user_id, topic_id, key)
                            WHERE topic_id IS NOT NULL
                        DO UPDATE SET
                            value = EXCLUDED.value,
                            source = EXCLUDED.source,
                            updated_at = now()
                        WHERE {_POSTGRES_EXISTING_SOURCE_PRIORITY}
                              <= {_POSTGRES_INCOMING_SOURCE_PRIORITY}
                    """
                    ),
                    {
                        "user_id": user_id,
                        "topic_id": topic_id,
                        "key": key,
                        "value": json_value,
                        "source": source,
                    },
                )
            else:
                # Global fact (topic_id IS NULL)
                await session.execute(
                    text(
                        f"""
                        INSERT INTO facts (user_id, topic_id, key, value, source)
                        VALUES ({_USER_ID_SUB}, NULL, :key, CAST(:value AS jsonb), :source)
                        ON CONFLICT (user_id, key)
                            WHERE topic_id IS NULL
                        DO UPDATE SET
                            value = EXCLUDED.value,
                            source = EXCLUDED.source,
                            updated_at = now()
                        WHERE {_POSTGRES_EXISTING_SOURCE_PRIORITY}
                              <= {_POSTGRES_INCOMING_SOURCE_PRIORITY}
                    """
                    ),
                    {
                        "user_id": user_id,
                        "key": key,
                        "value": json_value,
                        "source": source,
                    },
                )

    async def get_facts(self, user_id: str, topic_id: str | None = None) -> dict[str, Any]:
        """Get facts: global + topic-scoped (if provided)."""
        async with self._session() as session:
            if topic_id:
                result = await session.execute(
                    text(
                        f"""
                        SELECT key, value, topic_id FROM facts
                        WHERE user_id = {_USER_ID_SUB}
                          AND (topic_id IS NULL OR topic_id = :topic_id)
                        ORDER BY updated_at DESC
                    """
                    ),
                    {"user_id": user_id, "topic_id": topic_id},
                )
            else:
                result = await session.execute(
                    text(
                        f"""
                        SELECT key, value FROM facts
                        WHERE user_id = {_USER_ID_SUB} AND topic_id IS NULL
                        ORDER BY updated_at DESC
                    """
                    ),
                    {"user_id": user_id},
                )
            rows = result.fetchall()
        if not topic_id:
            return {r.key: r.value for r in rows}
        return _merge_scoped_postgres_facts(rows)

    # --- Summaries (SummaryStore) ---

    async def save_summary(
        self,
        user_id: str,
        topic_id: str,
        summary: str,
        messages_covered: int,
    ) -> None:
        """Save/update a rolling summary with versioning."""
        async with self._session(commit=True) as session:
            await session.execute(
                text(
                    f"""
                    INSERT INTO summaries (user_id, topic_id, summary, messages_covered, version)
                    VALUES ({_USER_ID_SUB}, :topic_id, :summary, :messages_covered, 1)
                    ON CONFLICT (user_id, topic_id)
                    DO UPDATE SET
                        summary = EXCLUDED.summary,
                        messages_covered = EXCLUDED.messages_covered,
                        version = summaries.version + 1,
                        updated_at = now()
                """
                ),
                {
                    "user_id": user_id,
                    "topic_id": topic_id,
                    "summary": summary,
                    "messages_covered": messages_covered,
                },
            )

    async def get_summary(self, user_id: str, topic_id: str) -> str | None:
        """Get the topic summary."""
        async with self._session() as session:
            result = await session.execute(
                text(
                    f"""
                    SELECT summary FROM summaries
                    WHERE user_id = {_USER_ID_SUB} AND topic_id = :topic_id
                """
                ),
                {"user_id": user_id, "topic_id": topic_id},
            )
            row = result.fetchone()
            return row.summary if row else None

    # --- Users (UserStore) ---

    async def ensure_user(self, external_id: str) -> str:
        """Create the user if needed and return external_id."""
        async with self._session(commit=True) as session:
            await session.execute(
                text(
                    """
                    INSERT INTO users (external_id)
                    VALUES (:external_id)
                    ON CONFLICT (external_id) DO NOTHING
                """
                ),
                {"external_id": external_id},
            )
        return external_id

    # --- Goals (GoalStore) ---

    async def save_goal(self, user_id: str, goal: GoalState) -> None:
        """Save or update a goal."""
        async with self._session(commit=True) as session:
            await session.execute(
                text(
                    f"""
                    INSERT INTO goals (user_id, topic_id, title, target_amount, current_amount,
                                       phase, is_main, plan, status)
                    VALUES (
                        {_USER_ID_SUB},
                        :topic_id, :title, :target_amount, :current_amount,
                        :phase, :is_main, CAST(:plan AS jsonb), 'active'
                    )
                    ON CONFLICT (user_id, topic_id) DO UPDATE SET
                        title = EXCLUDED.title,
                        target_amount = EXCLUDED.target_amount,
                        current_amount = EXCLUDED.current_amount,
                        phase = EXCLUDED.phase,
                        is_main = EXCLUDED.is_main,
                        plan = EXCLUDED.plan,
                        status = 'active'
                """
                ),
                {
                    "user_id": user_id,
                    "topic_id": goal.goal_id,
                    "title": goal.title,
                    "target_amount": goal.target_amount,
                    "current_amount": goal.current_amount,
                    "phase": goal.phase,
                    "is_main": goal.is_main,
                    "plan": json.dumps(goal.plan) if goal.plan else None,
                },
            )

    async def get_active_goal(self, user_id: str, topic_id: str) -> GoalState | None:
        """Get the active goal for the topic."""
        async with self._session() as session:
            result = await session.execute(
                text(
                    f"""
                    SELECT topic_id, title, target_amount, current_amount, phase, plan, is_main
                    FROM goals
                    WHERE user_id = {_USER_ID_SUB}
                      AND topic_id = :topic_id
                      AND status = 'active'
                    ORDER BY priority DESC, created_at DESC
                    LIMIT 1
                """
                ),
                {"user_id": user_id, "topic_id": topic_id},
            )
            row = result.fetchone()
            if not row:
                return None
            return GoalState(
                goal_id=row.topic_id,
                title=row.title,
                target_amount=row.target_amount,
                current_amount=row.current_amount,
                phase=row.phase,
                plan=row.plan,
                is_main=row.is_main,
            )

    # --- Session state (SessionStateStore - topics table) ---

    async def save_session_state(
        self,
        user_id: str,
        topic_id: str,
        role_id: str,
        active_skill_ids: list[str],
        prompt_hash: str = "",
        *,
        delegated_from: str | None = None,
        delegation_turn_count: int = 0,
        pending_delegation: str | None = None,
        delegation_summary: str | None = None,
    ) -> None:
        """Save session state for rehydration (topics table, §8.4)."""
        async with self._session(commit=True) as session:
            await session.execute(
                text(
                    f"""
                    INSERT INTO topics (user_id, topic_id, role_id, active_skill_ids, prompt_hash,
                                        delegated_from, delegation_turn_count, pending_delegation, delegation_summary)
                    VALUES ({_USER_ID_SUB}, :topic_id, :role_id, CAST(:skill_ids AS jsonb), :prompt_hash,
                            :delegated_from, :delegation_turn_count, :pending_delegation, :delegation_summary)
                    ON CONFLICT (user_id, topic_id)
                    DO UPDATE SET
                        role_id = EXCLUDED.role_id,
                        active_skill_ids = EXCLUDED.active_skill_ids,
                        prompt_hash = EXCLUDED.prompt_hash,
                        delegated_from = EXCLUDED.delegated_from,
                        delegation_turn_count = EXCLUDED.delegation_turn_count,
                        pending_delegation = EXCLUDED.pending_delegation,
                        delegation_summary = EXCLUDED.delegation_summary,
                        updated_at = now()
                """
                ),
                {
                    "user_id": user_id,
                    "topic_id": topic_id,
                    "role_id": role_id,
                    "skill_ids": json.dumps(active_skill_ids),
                    "prompt_hash": prompt_hash,
                    "delegated_from": delegated_from,
                    "delegation_turn_count": delegation_turn_count,
                    "pending_delegation": pending_delegation,
                    "delegation_summary": delegation_summary,
                },
            )

    async def get_session_state(self, user_id: str, topic_id: str) -> dict[str, Any] | None:
        """Get session state (from the topics table)."""
        async with self._session() as session:
            result = await session.execute(
                text(
                    f"""
                    SELECT role_id, active_skill_ids, title, COALESCE(prompt_hash, '') as prompt_hash,
                           delegated_from, COALESCE(delegation_turn_count, 0) as delegation_turn_count,
                           pending_delegation, delegation_summary
                    FROM topics
                    WHERE user_id = {_USER_ID_SUB} AND topic_id = :topic_id
                """
                ),
                {"user_id": user_id, "topic_id": topic_id},
            )
            row = result.fetchone()
            if not row:
                return None
            return {
                "role_id": row.role_id,
                "active_skill_ids": row.active_skill_ids or [],
                "title": row.title,
                "prompt_hash": row.prompt_hash,
                "delegated_from": row.delegated_from,
                "delegation_turn_count": row.delegation_turn_count or 0,
                "pending_delegation": row.pending_delegation,
                "delegation_summary": row.delegation_summary,
            }

    # --- Profile (UserStore) ---

    async def get_user_profile(self, user_id: str) -> UserProfile:
        """Get the profile with global facts."""
        facts = await self.get_facts(user_id, topic_id=None)
        return UserProfile(user_id=user_id, facts=facts)

    # --- Phase state ---

    async def save_phase_state(
        self,
        user_id: str,
        phase: str,
        notes: str = "",
    ) -> None:
        """Save or update the user's current phase."""
        async with self._session(commit=True) as session:
            await session.execute(
                text(
                    f"""
                    INSERT INTO phase_state (user_id, phase, notes)
                    VALUES ({_USER_ID_SUB}, :phase, :notes)
                    ON CONFLICT (user_id)
                    DO UPDATE SET
                        phase = EXCLUDED.phase,
                        notes = EXCLUDED.notes,
                        updated_at = now()
                """
                ),
                {"user_id": user_id, "phase": phase, "notes": notes},
            )

    async def get_phase_state(self, user_id: str) -> PhaseState | None:
        """Get the user's current phase."""
        async with self._session() as session:
            result = await session.execute(
                text(
                    f"""
                    SELECT phase, notes FROM phase_state
                    WHERE user_id = {_USER_ID_SUB}
                """
                ),
                {"user_id": user_id},
            )
            row = result.fetchone()
            if not row:
                return None
            return PhaseState(
                user_id=user_id,
                phase=row.phase,
                notes=row.notes or "",
            )

    # --- Tool events (§9.1) ---

    async def save_tool_event(
        self,
        user_id: str,
        event: ToolEvent,
    ) -> None:
        """Save a tool invocation event."""
        async with self._session(commit=True) as session:
            await session.execute(
                text(
                    f"""
                    INSERT INTO tool_events (user_id, topic_id, tool_name,
                                             input_json, output_json, latency_ms)
                    VALUES ({_USER_ID_SUB}, :topic_id, :tool_name,
                            CAST(:input_json AS jsonb), CAST(:output_json AS jsonb), :latency_ms)
                """
                ),
                {
                    "user_id": user_id,
                    "topic_id": event.topic_id,
                    "tool_name": event.tool_name,
                    "input_json": _json_or_none(event.input_json),
                    "output_json": _json_or_none(event.output_json),
                    "latency_ms": event.latency_ms,
                },
            )


def _json_or_none(value: Any) -> str | None:
    """Serialize a value to a JSON string or return None."""
    if value is None:
        return None
    return json.dumps(value)


def _merge_scoped_postgres_facts(rows: Sequence[Any]) -> dict[str, Any]:
    """Merge global + topic rows so topic-scoped values override global ones."""
    merged: dict[str, Any] = {}
    global_rows = [row for row in rows if getattr(row, "topic_id", None) is None]
    topic_rows = [row for row in rows if getattr(row, "topic_id", None) is not None]
    for row in global_rows + topic_rows:
        merged[row.key] = row.value
    return merged



"""Pre-LLM Input Filters — applied to messages and system_prompt before each turn.

Provides:
- InputFilter: runtime-checkable protocol for all filters
- MaxTokensFilter: truncates oldest messages to fit within token budget
- SystemPromptInjector: appends or prepends extra text to system prompt
"""

from __future__ import annotations

from typing import Protocol, runtime_checkable

from cognitia.runtime.types import Message


@runtime_checkable
class InputFilter(Protocol):
    """Filter applied to messages/system_prompt before each LLM call."""

    async def filter(
        self, messages: list[Message], system_prompt: str
    ) -> tuple[list[Message], str]:
        """Return (filtered_messages, filtered_system_prompt)."""
        ...


class MaxTokensFilter:
    """Truncates oldest messages to fit within max_tokens estimate.

    System prompt tokens are counted but never truncated.
    The last message is always preserved (even if it exceeds budget alone).
    """

    def __init__(
        self, max_tokens: int = 100_000, chars_per_token: float = 4.0
    ) -> None:
        self._max_tokens = max_tokens
        self._chars_per_token = chars_per_token

    def _estimate_tokens(self, text: str) -> int:
        return int(len(text) / self._chars_per_token)

    async def filter(
        self, messages: list[Message], system_prompt: str
    ) -> tuple[list[Message], str]:
        if not messages:
            return messages, system_prompt

        budget = self._max_tokens - self._estimate_tokens(system_prompt)
        if budget <= 0:
            # System prompt alone exceeds budget; keep only last message
            return [messages[-1]], system_prompt

        # Calculate total message tokens
        msg_tokens = [self._estimate_tokens(m.content) for m in messages]
        total = sum(msg_tokens)

        if total <= budget:
            return messages, system_prompt

        # Trim from the front (oldest), always keep at least the last message
        trimmed = list(messages)
        trimmed_tokens = list(msg_tokens)

        while len(trimmed) > 1 and sum(trimmed_tokens) > budget:
            trimmed.pop(0)
            trimmed_tokens.pop(0)

        return trimmed, system_prompt


class SystemPromptInjector:
    """Appends or prepends extra text to the system prompt.

    Args:
        extra_text: Text to inject.
        position: "append" (default) or "prepend".
    """

    def __init__(self, extra_text: str, position: str = "append") -> None:
        self._extra_text = extra_text
        self._position = position

    async def filter(
        self, messages: list[Message], system_prompt: str
    ) -> tuple[list[Message], str]:
        if not system_prompt:
            return messages, self._extra_text

        if self._position == "prepend":
            return messages, f"{self._extra_text}\n{system_prompt}"

        return messages, f"{system_prompt}\n{self._extra_text}"



"""TddCodeVerifier - CodeVerifier implementation respecting CodingStandardsConfig.

Disabled checks return VerificationStatus.SKIP automatically.
"""

from __future__ import annotations

from cognitia.orchestration.code_verifier import CommandResult, CommandRunner
from cognitia.orchestration.coding_standards import CodingStandardsConfig
from cognitia.orchestration.verification_types import (
    CheckDetail,
    VerificationResult,
    VerificationStatus,
)


class TddCodeVerifier:
    """CodeVerifier implementation - respects CodingStandardsConfig.

  Disabled check -> auto VerificationStatus.SKIP.
  """

    def __init__(self, config: CodingStandardsConfig, runner: CommandRunner) -> None:
        self._config = config
        self._runner = runner

    async def verify_contracts(self) -> VerificationResult:
        """Verify contract tests pass."""
        if not self._config.solid_enabled:
            return self._skip("Contract verification disabled (SOLID off)")
        result = await self._runner.run("pytest -m contract -q")
        return self._from_command(result, "contracts")

    async def verify_tests_substantive(self) -> VerificationResult:
        """Verify tests are substantive (not just pass/trivial)."""
        if not self._config.tdd_enabled:
            return self._skip("TDD disabled")
        result = await self._runner.run("pytest -q")
        return self._from_command(result, "tests_substantive")

    async def verify_tests_before_code(self) -> VerificationResult:
        """Verify tests were written before implementation (heuristic)."""
        if not self._config.tdd_enabled:
            return self._skip("TDD disabled")
        result = await self._runner.run("git log --oneline -5")
        return self._from_command(result, "tests_before_code")

    async def verify_linters(self) -> VerificationResult:
        """Run linters (ruff check)."""
        result = await self._runner.run("ruff check .")
        status = VerificationStatus.PASS if result.exit_code == 0 else VerificationStatus.FAIL
        return VerificationResult(
            status=status,
            checks=(
                CheckDetail(name="ruff", status=status, message=result.stdout or result.stderr),
            ),
            summary=result.stdout or result.stderr,
        )

    async def verify_coverage(self, min_pct: int = 85) -> VerificationResult:
        """Run tests with coverage check."""
        if not self._config.tdd_enabled:
            return self._skip("TDD disabled")
        effective_min = max(min_pct, self._config.min_coverage_pct)
        result = await self._runner.run(f"pytest --cov --cov-fail-under={effective_min}")
        status = VerificationStatus.PASS if result.exit_code == 0 else VerificationStatus.FAIL
        return VerificationResult(
            status=status,
            checks=(
                CheckDetail(
                    name="coverage",
                    status=status,
                    message=f"min={effective_min}%",
                ),
            ),
            summary=result.stdout or result.stderr,
        )

    def _skip(self, reason: str) -> VerificationResult:
        return VerificationResult(status=VerificationStatus.SKIP, summary=reason)

    def _from_command(self, result: CommandResult, check_name: str) -> VerificationResult:
        status = VerificationStatus.PASS if result.exit_code == 0 else VerificationStatus.FAIL
        return VerificationResult(
            status=status,
            checks=(
                CheckDetail(
                    name=check_name,
                    status=status,
                    message=result.stdout or result.stderr,
                ),
            ),
            summary=result.stdout or result.stderr,
        )



"""Canonical built-in tool names and aliases - single source of truth.

Used by both ThinRuntime and DeepAgents runtime for built-in tool
identification, alias resolution, and portable/native mode filtering.
"""

from __future__ import annotations

BUILTIN_TOOL_NAMES: frozenset[str] = frozenset(
    {
        "read_file",
        "write_file",
        "edit_file",
        "ls",
        "glob",
        "grep",
        "execute",
        "write_todos",
        "task",
    }
)

BUILTIN_ALIASES: dict[str, str] = {
    "TodoRead": "write_todos",
    "TodoWrite": "write_todos",
    "LS": "ls",
    "Read": "read_file",
    "Write": "write_file",
    "Edit": "edit_file",
    "MultiEdit": "edit_file",
    "Glob": "glob",
    "Grep": "grep",
    "Bash": "execute",
    "Task": "task",
}



"""CLI Agent Runtime -- subprocess-based runtime for external CLI agents.

Runs an external CLI process, feeds it prompt via stdin, and parses
NDJSON output from stdout into RuntimeEvent stream.
"""

from __future__ import annotations

import asyncio
import logging
import os
from collections.abc import AsyncIterator
from typing import Any

from cognitia.runtime.cli.parser import ClaudeNdjsonParser, GenericNdjsonParser, NdjsonParser
from cognitia.runtime.cli.types import CliConfig
from cognitia.runtime.types import (
    Message,
    RuntimeConfig,
    RuntimeErrorData,
    RuntimeEvent,
    ToolSpec,
)

logger = logging.getLogger(__name__)


def _build_stdin_payload(messages: list[Message], system_prompt: str) -> str:
    """Serialize system instructions and conversation for CLI stdin."""
    sections: list[str] = []
    if system_prompt.strip():
        sections.append(f"System instructions:\n{system_prompt}")

    conversation = "\n".join(f"{m.role}: {m.content}" for m in messages)
    sections.append(f"Conversation:\n{conversation}")
    return "\n\n".join(sections)


def _is_claude_command(command: list[str]) -> bool:
    """Is claude command."""
    if not command:
        return False
    return os.path.basename(command[0]) == "claude"


def _normalize_claude_command(command: list[str], output_format: str) -> list[str]:
    """Normalize claude command."""
    if not _is_claude_command(command):
        return list(command)

    normalized = [
        "--output-format" if token == "--output" else token
        for token in command
    ]


    prompt_placeholder = normalized[-1] == "-" if normalized else False
    core = normalized[:-1] if prompt_placeholder else normalized[:]

    if "--print" not in core and "-p" not in core:
        core.append("--print")

    if "--output-format" not in core:
        core.extend(["--output-format", output_format])

    if output_format == "stream-json" and "--verbose" not in core and "-v" not in core:
        core.append("--verbose")

    if prompt_placeholder:
        core.append("-")
    return core


class CliAgentRuntime:
    """Run external CLI agents via subprocess + NDJSON parsing.

  Implements AgentRuntime protocol. Parser is auto-selected based on
  the command name (``Claude`` -> ClaudeNdJSONParser, else Generic).
  """

    def __init__(
        self,
        config: RuntimeConfig,
        cli_config: CliConfig | None = None,
        parser: NdjsonParser | None = None,
    ) -> None:
        self._config = config
        self._cli_config = cli_config or CliConfig(
            command=["claude", "--print", "--verbose", "--output-format", "stream-json", "-"]
        )
        if parser is not None:
            self._parser = parser
        elif _is_claude_command(self._cli_config.command):
            self._parser = ClaudeNdjsonParser()
        else:
            self._parser = GenericNdjsonParser()
        self._process: asyncio.subprocess.Process | None = None
        self._cancel_requested = False

    async def run(  # noqa: C901
        self,
        *,
        messages: list[Message],
        system_prompt: str,
        active_tools: list[ToolSpec],
        config: RuntimeConfig | None = None,
        mode_hint: str | None = None,
    ) -> AsyncIterator[RuntimeEvent]:
        """Execute CLI agent and yield RuntimeEvents from its NDJSON output."""
        prompt = _build_stdin_payload(messages, system_prompt)
        cmd = _normalize_claude_command(
            self._cli_config.command,
            self._cli_config.output_format,
        )
        env = dict(os.environ)
        env.update(self._cli_config.env)

        total_bytes = 0
        max_bytes = self._cli_config.max_output_bytes
        timeout = self._cli_config.timeout_seconds
        final_seen = False
        self._cancel_requested = False

        try:
            self._process = await asyncio.create_subprocess_exec(
                *cmd,
                stdin=asyncio.subprocess.PIPE,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
                env=env,
            )

            if self._process.stdin:
                self._process.stdin.write(prompt.encode())
                await self._process.stdin.drain()
                self._process.stdin.close()

            async def _read_and_parse() -> AsyncIterator[RuntimeEvent]:
                nonlocal total_bytes, final_seen
                if not self._process or not self._process.stdout:
                    return
                async for raw_line in self._process.stdout:
                    total_bytes += len(raw_line)
                    if total_bytes > max_bytes:
                        self._terminate_process()
                        yield RuntimeEvent.error(
                            RuntimeErrorData(
                                kind="budget_exceeded",
                                message=f"Output exceeded {max_bytes} bytes",
                            )
                        )
                        return
                    line = raw_line.decode("utf-8", errors="replace").rstrip()
                    if not line:
                        continue
                    event = self._parser.parse_line(line)
                    if event is not None:
                        if event.is_final:
                            final_seen = True
                        yield event

            try:
                async with asyncio.timeout(timeout):
                    async for event in _read_and_parse():
                        yield event
            except TimeoutError:
                self._terminate_process()
                yield RuntimeEvent.error(
                    RuntimeErrorData(
                        kind="mcp_timeout",
                        message=f"CLI process timed out after {timeout}s",
                        recoverable=False,
                    )
                )
                return

            await self._process.wait()
            if self._cancel_requested:
                yield RuntimeEvent.error(
                    RuntimeErrorData(
                        kind="cancelled",
                        message="CLI process cancelled",
                        recoverable=False,
                    )
                )
            elif self._process.returncode and self._process.returncode != 0:
                stderr_data = b""
                if self._process.stderr:
                    stderr_data = await self._process.stderr.read()
                yield RuntimeEvent.error(
                    RuntimeErrorData(
                        kind="runtime_crash",
                        message=(
                            f"Process exited with code {self._process.returncode}: "
                            f"{stderr_data.decode(errors='replace')}"
                        ),
                    )
                )
            elif not final_seen:
                yield RuntimeEvent.error(
                    RuntimeErrorData(
                        kind="bad_model_output",
                        message="CLI process exited without a final NDJSON event",
                    )
                )

        except Exception:
            logger.exception("CliAgentRuntime.run() failed")
            yield RuntimeEvent.error(
                RuntimeErrorData(kind="runtime_crash", message="Unexpected error in CLI runtime")
            )

    def cancel(self) -> None:
        """Request cooperative cancellation of the current subprocess."""
        if self._process and self._process.returncode is None:
            self._cancel_requested = True
        self._terminate_process()

    async def cleanup(self) -> None:
        """Release subprocess resources."""
        self._terminate_process()
        if self._process and self._process.returncode is None:
            try:
                await asyncio.wait_for(self._process.wait(), timeout=5.0)
            except TimeoutError:
                self._process.kill()
        self._cancel_requested = False

    def _terminate_process(self) -> None:
        """Send SIGTERM to subprocess if still running."""
        if self._process and self._process.returncode is None:
            try:
                self._process.terminate()
            except ProcessLookupError:
                pass

    async def __aenter__(self) -> CliAgentRuntime:
        return self

    async def __aexit__(self, *exc: Any) -> None:
        await self.cleanup()



"""LLM client functions for ThinRuntime."""

from __future__ import annotations

import asyncio
from dataclasses import dataclass
import logging
from collections.abc import AsyncIterator, Callable
from typing import Any

from cognitia.runtime.provider_resolver import resolve_provider
from cognitia.runtime.thin.errors import ThinLlmError, provider_runtime_crash
from cognitia.runtime.thin.llm_providers import get_cached_adapter
from cognitia.runtime.types import RuntimeConfig, RuntimeErrorData

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class BufferedLlmAttempt:
    """Buffered Llm Attempt implementation."""

    raw: str
    chunks: list[str]
    used_stream: bool


async def try_stream_llm_call(
    llm_call: Callable[..., Any],
    lm_messages: list[dict[str, str]],
    prompt: str,
) -> tuple[list[str], str] | None:
    """Try stream llm call."""
    try:
        result = await llm_call(lm_messages, prompt, stream=True)
    except TypeError:
        # LLM not supports stream kwarg
        return None

    if isinstance(result, str):

        return [result], result

    if not hasattr(result, "__aiter__"):
        return None

    chunks: list[str] = []
    async for chunk in result:
        chunks.append(chunk)
    return chunks, "".join(chunks)


async def run_buffered_llm_call(
    llm_call: Callable[..., Any],
    lm_messages: list[dict[str, str]],
    prompt: str,
    *,
    retry_policy: Any | None = None,
    cancellation_token: Any | None = None,
    on_retry: Callable[[int, float], None] | None = None,
) -> BufferedLlmAttempt:
    """Run buffered llm call."""
    attempt = 0
    while True:
        try:
            _raise_if_cancelled(cancellation_token)
            try:
                result = await llm_call(lm_messages, prompt, stream=True)
            except TypeError:
                raw = await llm_call(lm_messages, prompt)
                return BufferedLlmAttempt(raw=raw, chunks=[], used_stream=False)

            if isinstance(result, str):
                return BufferedLlmAttempt(raw=result, chunks=[result], used_stream=True)

            if hasattr(result, "__aiter__"):
                chunks: list[str] = []
                async for chunk in result:
                    chunks.append(chunk)
                return BufferedLlmAttempt(raw="".join(chunks), chunks=chunks, used_stream=True)

            raw = await llm_call(lm_messages, prompt)
            return BufferedLlmAttempt(raw=raw, chunks=[], used_stream=False)
        except ThinLlmError as exc:
            if retry_policy is None or exc.error.kind == "cancelled":
                raise

            should_retry, delay = retry_policy.should_retry(exc, attempt)
            if not should_retry:
                raise

            logger.info(
                "Retry attempt %d, delay %.1fs during buffered stream: %s",
                attempt + 1,
                delay,
                exc,
            )
            if on_retry is not None:
                on_retry(attempt + 1, delay)
            await _sleep_with_cancellation(delay, cancellation_token)
            attempt += 1


async def _stream_with_error_normalization(
    adapter: Any,
    provider: str,
    messages: list[dict[str, str]],
    system_prompt: str,
    **kwargs: Any,
) -> AsyncIterator[str]:
    """Wrap adapter.stream() so iteration-time failures become typed errors."""
    try:
        async for chunk in adapter.stream(messages, system_prompt, **kwargs):
            yield chunk
    except ThinLlmError:
        raise
    except Exception as exc:
        logger.error("Ошибка LLM API (%s)", provider, exc_info=True)
        raise provider_runtime_crash(provider, exc) from exc


async def default_llm_call(
    config: RuntimeConfig,
    messages: list[dict[str, str]],
    system_prompt: str,
    **kwargs: Any,
) -> str | AsyncIterator[str]:
    """Default llm call."""
    use_stream = kwargs.pop("stream", False)

    resolved = resolve_provider(config.model, base_url=config.base_url)
    logger.info(
        "LLM запрос: model=%s, provider=%s, sdk=%s, stream=%s",
        resolved.model_id,
        resolved.provider,
        resolved.sdk_type,
        use_stream,
    )

    try:
        adapter = get_cached_adapter(resolved)
    except ThinLlmError:
        raise
    except Exception as exc:
        logger.error("Ошибка инициализации LLM адаптера (%s)", resolved.provider, exc_info=True)
        raise provider_runtime_crash(resolved.provider, exc) from exc

    try:
        if use_stream:
            return _stream_with_error_normalization(
                adapter,
                resolved.provider,
                messages,
                system_prompt,
                **kwargs,
            )
        return await adapter.call(messages, system_prompt, **kwargs)
    except ThinLlmError:
        raise
    except Exception as exc:
        logger.error("Ошибка LLM API (%s)", resolved.provider, exc_info=True)
        raise provider_runtime_crash(resolved.provider, exc) from exc


def _raise_if_cancelled(token: Any | None) -> None:
    if token is None or not token.is_cancelled:
        return
    raise ThinLlmError(
        RuntimeErrorData(
            kind="cancelled",
            message="Operation cancelled",
            recoverable=False,
        )
    )


async def _sleep_with_cancellation(delay: float, token: Any | None) -> None:
    if delay <= 0:
        return
    if token is None:
        await asyncio.sleep(delay)
        return

    remaining = delay
    step = 0.05
    while remaining > 0:
        _raise_if_cancelled(token)
        wait_for = min(step, remaining)
        await asyncio.sleep(wait_for)
        remaining -= wait_for
    _raise_if_cancelled(token)



"""Helpers for portable structured output in Thin/DeepAgents runtimes."""

from __future__ import annotations

import json
from typing import Any, Protocol, cast

from cognitia.runtime.thin.parsers import strip_markdown_fences as _strip_markdown_fences


class _StructuredOutputModel(Protocol):
    """Structured Output Model protocol."""

    @classmethod
    def model_validate_json(cls, json_data: str) -> Any: ...

    @classmethod
    def model_json_schema(cls) -> dict[str, Any]: ...


def validate_structured_output(text: str, output_type: _StructuredOutputModel) -> Any:
    """Parse JSON from text and validate against a Pydantic model.

  Strips markdown fences before parsing. Returns a validated model instance.

  Raises:
    ValueError: if text is not valid JSON.
    pydantic.ValidationError: if JSON does not match the model schema.
  """
    cleaned = _strip_markdown_fences(text).strip()
    try:
        json.loads(cleaned)  # validate JSON syntax first
    except json.JSONDecodeError as exc:
        raise ValueError(f"Invalid JSON: {exc}") from exc
    return output_type.model_validate_json(cleaned)


def extract_pydantic_schema(output_type: _StructuredOutputModel) -> dict[str, Any]:
    """Extract JSON Schema dict from a Pydantic model class."""
    return output_type.model_json_schema()


def resolve_structured_output(
    text: str,
    output_format: dict[str, Any] | None,
    output_type: type | None,
) -> Any | None:
    """Resolve structured output: validate via Pydantic if output_type set, else extract JSON.

  Returns:
    - Pydantic model instance if output_type is set and validation succeeds
    - dict/list if output_format is set (no output_type) and JSON parses
    - None if no structured output configured

  Raises:
    ValueError: if output_type is set but text is not valid JSON
    pydantic.ValidationError: if output_type is set but JSON doesn't match schema
  """
    if output_type is not None:
        model = cast(_StructuredOutputModel, output_type)
        return validate_structured_output(text, model)
    return extract_structured_output(text, output_format)


def try_resolve_structured_output(
    text: str,
    output_format: dict[str, Any] | None,
    output_type: type | None,
) -> tuple[Any | None, str | None]:
    """Try to resolve structured output, returning (result, error_message).

  If output_type is set and validation fails, returns (None, error_str).
  If no output_type, falls back to extract_structured_output (never errors).
  On success returns (parsed_output, None).
  """
    if output_type is not None:
        try:
            model = cast(_StructuredOutputModel, output_type)
            parsed = validate_structured_output(text, model)
            return parsed, None
        except (ValueError, Exception) as exc:
            return None, str(exc)
    return extract_structured_output(text, output_format), None


def append_structured_output_instruction(
    system_prompt: str,
    output_format: dict[str, Any] | None,
    *,
    final_response_field: str | None = None,
) -> str:
    """Append structured output instruction."""
    schema = normalize_output_schema(output_format)
    if schema is None:
        return system_prompt

    schema_json = json.dumps(schema, ensure_ascii=False)
    if final_response_field is not None:
        instruction = (
            "\n\n## Structured output\n"
            "Your final response must be valid JSON conforming to the schema below.\n"
            f"Place the entire JSON in the `{final_response_field}` field, with no markdown or explanations.\n"
            f"Schema: {schema_json}"
        )
    else:
        instruction = (
            "\n\n## Structured output\n"
            "Your final response must be valid JSON only, conforming to the schema below.\n"
            "No markdown wrapping and no explanations.\n"
            f"Schema: {schema_json}"
        )
    return f"{system_prompt}{instruction}"


def extract_structured_output(
    text: str,
    output_format: dict[str, Any] | None,
) -> Any | None:
    """Extract structured output."""
    if not output_format:
        return None

    cleaned = _strip_markdown_fences(text).strip()
    if not cleaned:
        return None

    for candidate in (cleaned, _extract_first_json_value(cleaned)):
        if not candidate:
            continue
        try:
            return json.loads(candidate)
        except json.JSONDecodeError:
            continue
    return None


def normalize_output_schema(output_format: dict[str, Any] | None) -> dict[str, Any] | None:
    """Normalize output schema."""
    if not output_format:
        return None
    if output_format.get("type") == "json_schema":
        schema = output_format.get("schema")
        return schema if isinstance(schema, dict) else None
    return output_format


def _extract_first_json_value(text: str) -> str | None:
    decoder = json.JSONDecoder()
    for idx, char in enumerate(text):
        if char not in '{["-0123456789tfn':
            continue
        try:
            _, end = decoder.raw_decode(text[idx:])
            return text[idx : idx + end]
        except json.JSONDecodeError:
            continue
    return None



"""Thin Team module."""

from __future__ import annotations

import asyncio
import json
import uuid
from dataclasses import replace
from collections.abc import Callable
from datetime import UTC, datetime
from typing import Any

from cognitia.orchestration.base_team import BaseTeamOrchestrator
from cognitia.orchestration.message_bus import MessageBus
from cognitia.orchestration.message_tools import (
    SEND_MESSAGE_TOOL_SPEC,
    create_send_message_tool,
)
from cognitia.orchestration.team_types import (
    InternalTeamState,
    TeamConfig,
    compose_worker_task,
)
from cognitia.orchestration.thin_subagent import ThinSubagentOrchestrator
from cognitia.runtime.types import RuntimeConfig


async def _default_llm_call(
    messages: list[dict[str, Any]], system_prompt: str, **kwargs: Any
) -> str:
    """No-op LLM call -- inozin increases manandmal final from inno with not much delay.

  The delay is necessary so that workers do not finish immediately to the point of collapse
  pause/resume in test and production code.
  """
    await asyncio.sleep(0.1)
    return json.dumps({"type": "final", "final_message": "done"})


class ThinTeamOrchestrator(BaseTeamOrchestrator):
    """Thin Team Orchestrator implementation."""

    def __init__(
        self,
        *,
        llm_call: Callable[..., Any] | None = None,
        local_tools: dict[str, Callable[..., Any]] | None = None,
        mcp_servers: dict[str, Any] | None = None,
        runtime_config: RuntimeConfig | None = None,
        max_concurrent: int = 8,
        sub_orchestrator: ThinSubagentOrchestrator | None = None,
    ) -> None:
        if sub_orchestrator is not None:
            sub_orch = sub_orchestrator
        else:
            effective_llm_call = llm_call if llm_call is not None else _default_llm_call
            sub_orch = ThinSubagentOrchestrator(
                max_concurrent=max_concurrent,
                llm_call=effective_llm_call,
                local_tools=local_tools,
                mcp_servers=mcp_servers,
                runtime_config=runtime_config,
            )
        super().__init__(sub_orch)

    async def start(self, config: TeamConfig, task: str) -> str:
        """Start."""
        team_id = str(uuid.uuid4())
        worker_ids: dict[str, str] = {}
        bus = MessageBus()

        send_message_name = SEND_MESSAGE_TOOL_SPEC.name
        worker_specs = [
            replace(
                spec,
                tools=(
                    spec.tools
                    if any(tool.name == send_message_name for tool in spec.tools)
                    else [*spec.tools, SEND_MESSAGE_TOOL_SPEC]
                ),
            )
            for spec in config.worker_specs[: config.max_workers]
        ]
        worker_names = [s.name for s in worker_specs]

        # Regandstrandruem send_message tool - odandn executor for inseh workers
        # executor sends args dict with from_agent (optional, fallback on sender)
        assert isinstance(self._sub_orch, ThinSubagentOrchestrator)
        self._sub_orch.register_tool(
            "send_message",
            create_send_message_tool(
                bus,
                sender_agent_id="team",
                team_members=worker_names,
            ),
        )

        for spec in worker_specs:
            worker_task = compose_worker_task(
                config=config, worker_name=spec.name, task=task
            )
            agent_id = await self._sub_orch.spawn(spec, worker_task)
            worker_ids[spec.name] = agent_id

        self._teams[team_id] = InternalTeamState(
            config=config,
            worker_ids=worker_ids,
            started_at=datetime.now(tz=UTC),
            bus=bus,
            task=task,
        )
        return team_id

    def _resolve_worker(
        self, state: InternalTeamState, agent_id: str
    ) -> tuple[str, str] | None:
        """Resolve agent_id to (worker_name, actual_agent_uuid).

    Accepts both worker name (from status.workers keys) and UUID (internal).
    Returns None if not found.
    """
        # Try as UUID first
        by_uuid = next(
            (name for name, cur_id in state.worker_ids.items() if cur_id == agent_id),
            None,
        )
        if by_uuid is not None:
            return by_uuid, agent_id

        # Try as worker name
        if agent_id in state.worker_ids:
            return agent_id, state.worker_ids[agent_id]

        return None

    async def pause_agent(self, team_id: str, agent_id: str) -> None:
        """Pause agent."""
        state = self._teams.get(team_id)
        if not state:
            return
        resolved = self._resolve_worker(state, agent_id)
        if resolved is None:
            return
        worker_name, actual_uuid = resolved
        await self._sub_orch.cancel(actual_uuid)
        state.paused_workers.add(worker_name)

    async def resume_agent(self, team_id: str, agent_id: str) -> None:
        """Resume agent."""
        state = self._teams.get(team_id)
        if not state:
            return
        resolved = self._resolve_worker(state, agent_id)
        if resolved is None:
            return
        worker_name, _actual_uuid = resolved
        if worker_name not in state.paused_workers:
            return

        spec = next((s for s in state.config.worker_specs if s.name == worker_name), None)
        if spec is None:
            return

        worker_task = compose_worker_task(
            config=state.config,
            worker_name=worker_name,
            task=state.task,
        )
        new_agent_id = await self._sub_orch.spawn(spec, worker_task)
        state.worker_ids[worker_name] = new_agent_id
        state.paused_workers.discard(worker_name)



"""SummaryGenerator - generate a rolling summary from message history.

MVP: template-based (no LLM call).
Can later be replaced with an LLM-based implementation via the same Protocol.
"""

from __future__ import annotations

from cognitia.memory.types import MemoryMessage


class TemplateSummaryGenerator:
    """Template-based summary generator (KISS for the MVP).

    Takes the last N messages, truncates long ones, and builds
    a short textual summary in the format:
    - [user]: text
    - [assistant]: text
    """

    def __init__(
        self,
        max_messages: int = 20,
        max_message_chars: int = 200,
    ) -> None:
        self._max_messages = max_messages
        self._max_message_chars = max_message_chars

    def summarize(self, messages: list[MemoryMessage]) -> str:
        """Generate a summary from a list of messages.

        Args:
            messages: List of messages (oldest to newest).

        Returns:
            Summary text or an empty string if there are no messages.
        """
        if not messages:
            return ""

        # Take only the latest N messages
        recent = messages[-self._max_messages :]

        lines: list[str] = []
        for msg in recent:
            content = msg.content
            if len(content) > self._max_message_chars:
                content = content[: self._max_message_chars] + "..."
            role_label = "user" if msg.role == "user" else "assistant"
            lines.append(f"- [{role_label}]: {content}")

        return "\n".join(lines)



"""Verifier Port module."""

from __future__ import annotations

from typing import Protocol


class VerifierPort(Protocol):
    """Verify execution output. Returns (passed, feedback)."""

    async def verify(self, output: str) -> tuple[bool, str]: ...


class ExecutorPort(Protocol):
    """Execute a goal/task. Returns output string."""

    async def execute(self, goal: str) -> str: ...
